import logging

from ..common import Object
from .base_command import BaseCommand


class Settings(BaseCommand):
    """
    CTERA Portal Settings APIs

    :ivar cterasdk.core.settings.GlobalSettings global_settings: Object holding the Portal Global Settings APIs
    """

    def __init__(self, core):
        super().__init__(core)
        self.global_settings = GlobalSettings(core)
        self.portal = PortalSettings(core)


class PortalSettings(BaseCommand):
    """
    Virtual Portal Settings APIs
    """
    def get(self):
        if self.session().in_tenant_context():
            return self._core.api.execute('', 'getSettings').settings
        return self._core.api.get('/settings/defaultPortalSettings')

    def use_global_settings(self):
        return self.update()

    def update(self, settings=None):
        """
        Update Portal Settings

        :param cterasdk.common.object.Objcet settings: Settings, defaults to using the global settings.
        """
        if not self.session().in_tenant_context():
            self._core.api.put('/settings/defaultPortalSettings', settings)
        else:
            param = PortalSettings._create_settings_parameter()
            if settings:
                param.fromSystem = False
                param.settings = settings
            self._core.api.execute('', 'setSettings', param)

    @staticmethod
    def _create_settings_parameter():
        param = Object()
        param._classname = 'SettingsParam'  # pylint: disable=protected-access
        param.fromSystem = True
        return param


class GlobalSettings(BaseCommand):
    """
    Global Settings APIs
    """

    def get_timezone(self):
        """
        Get timezone
        """
        return self._core.api.get('/settings/timezone')

    def set_timezone(self, timezone):
        """
        Set timezone

        :param str timezone: Timezone
        """
        logging.getLogger('cterasdk.core').info('Updating timezone. %s', {'timezone': timezone})
        response = self._core.api.put('/settings/timezone', timezone)
        logging.getLogger('cterasdk.core').info('Updated timezone. %s', {'timezone': timezone})
        return response



from ...clients.synchronous import clients
from ..services import Management
from ..endpoints import EndpointBuilder

from .. import authenticators

from ...lib.session.edge import Session

from ...edge import afp
from ...edge import aio
from ...edge import array
from ...edge import audit
from ...edge import backup
from ...edge import cache
from ...edge import cli
from ...edge import config
from ...edge import connection
from ...edge import ctera_migrate
from ...edge import dedup
from ...edge import directoryservice
from ...edge import drive
from ...edge import files
from ...edge import firmware
from ...edge import ftp
from ...edge import groups
from ...edge import licenses
from ...edge import login
from ...edge import logs
from ...edge import mail
from ...edge import network
from ...edge import nfs
from ...edge import ntp
from ...edge import power
from ...edge import remote
from ...edge import rsync
from ...edge import ransom_protect
from ...edge import services
from ...edge import shares
from ...edge import shell
from ...edge import smb
from ...edge import snmp
from ...edge import ssh
from ...edge import ssl
from ...edge import support
from ...edge import sync
from ...edge import syslog
from ...edge import taskmgr
from ...edge import telnet
from ...edge import timezone
from ...edge import users
from ...edge import volumes


class Clients:

    def __init__(self, edge, Portal):
        if Portal:
            edge._Portal = Portal
            edge.default.close()
            edge._ctera_session.start_remote_session(Portal.session())
            self.api = Portal.default.clone(clients.API, EndpointBuilder.new(edge.base), authenticator=lambda *_: True)
        else:
            self.migrate = edge.default.clone(clients.Migrate, EndpointBuilder.new(edge.base, '/migration/rest/v1'))
            self.api = edge.default.clone(clients.API, EndpointBuilder.new(edge.base, '/admingui/api'))
            self.io = IO(edge)


class IO:

    def __init__(self, edge):
        self._edge = edge
        self._webdav = edge.default.clone(clients.WebDAV, EndpointBuilder.new(edge.base, '/localFiles'))

    @property
    def download(self):
        return self._webdav.download

    @property
    def download_zip(self):
        return self._edge.default.form_data  # pylint: disable=protected-access

    @property
    def upload(self):
        return self._edge.default.form_data  # pylint: disable=protected-access

    @property
    def mkdir(self):
        return self._webdav.mkcol

    @property
    def copy(self):
        return self._webdav.copy

    @property
    def move(self):
        return self._webdav.move

    @property
    def delete(self):
        return self._webdav.delete


class Edge(Management):  # pylint: disable=too-many-instance-attributes

    def __init__(self, host=None, port=None, https=True, Portal=None, *, base=None):
        super().__init__(host, port, https, base=base)
        self._ctera_session = Session(self.host())
        self._ctera_clients = Clients(self, Portal)

        self.afp = afp.AFP(self)
        self.aio = aio.AIO(self)
        self.array = array.Array(self)
        self.audit = audit.Audit(self)
        self.backup = backup.Backup(self)
        self.cache = cache.Cache(self)
        self.cli = cli.CLI(self)
        self.config = config.Config(self)
        self.ctera_migrate = ctera_migrate.CTERAMigrate(self)
        self.dedup = dedup.Dedup(self)
        self.directoryservice = directoryservice.DirectoryService(self)
        self.drive = drive.Drive(self)
        self.files = files.FileBrowser(self)
        self.firmware = firmware.Firmware(self)
        self.ftp = ftp.FTP(self)
        self.groups = groups.Groups(self)
        self.licenses = licenses.Licenses(self)
        self.logs = logs.Logs(self)
        self.mail = mail.Mail(self)
        self.network = network.Network(self)
        self.nfs = nfs.NFS(self)
        self.ntp = ntp.NTP(self)
        self.power = power.Power(self)
        self.ransom_protect = ransom_protect.RansomProtect(self)
        self.rsync = rsync.RSync(self)
        self.services = services.Services(self)
        self.shares = shares.Shares(self)
        self.shell = shell.Shell(self)
        self.smb = smb.SMB(self)
        self.snmp = snmp.SNMP(self)
        self.ssh = ssh.SSH(self)
        self.ssl = ssl.SSL(self)
        self.support = support.Support(self)
        self.sync = sync.Sync(self)
        self.syslog = syslog.Syslog(self)
        self.tasks = taskmgr.Tasks(self)
        self.telnet = telnet.Telnet(self)
        self.timezone = timezone.Timezone(self)
        self.users = users.Users(self)
        self.volumes = volumes.Volumes(self)

    def _after_login(self):
        self.ssl = ssl.initialize(self)

    @property
    def migrate(self):
        return self.clients.migrate

    @property
    def api(self):
        return self.clients.api

    @property
    def io(self):
        return self.clients.io

    @property
    def _session_id_key(self):
        return '_cteraSessionId_'

    def _authenticator(self, url):
        return authenticators.edge(self.session(), url)

    @property
    def _login_object(self):
        return login.Login(self)

    @property
    def initialized(self):
        return not self._login_object.info().isfirstlogin

    def test(self):
        return connection.test(self)

    def sso(self, ticket):
        """ Login using Single Sign On"""
        self._login_object.sso(ticket)
        self.session().start_session(self)

    def remote_access(self):
        return remote.remote_access(self, self._Portal)

    @property
    def _omit_fields(self):
        return super()._omit_fields + ['afp', 'aio', 'array', 'audit', 'backup', 'cache', 'cli', 'config', 'ctera_migrate', 'dedup',
                                       'directoryservice', 'drive', 'files', 'firmware', 'ftp', 'groups', 'licenses', 'logs', 'mail',
                                       'network', 'nfs', 'ntp', 'power', 'ransom_protect', 'rsync', 'services', 'shares', 'shell',
                                       'smb', 'snmp', 'ssh', 'ssl', 'support', 'sync', 'syslog', 'tasks', 'telnet', 'timezone',
                                       'users', 'volumes']



import logging

from ..exceptions import CTERAException
from .base_command import BaseCommand


class Login(BaseCommand):

    def info(self):
        """
        Get login info
        """
        return self._edge.api.get('/nosession/logininfo')

    def login(self, username, password):
        host = self._edge.host()
        try:
            self._edge.api.form_data('/login', {'username': username, 'password': password})
            logging.getLogger('cterasdk.edge').info("User logged in. %s", {'host': host, 'user': username})
            self._edge.ctera_migrate.login()
        except CTERAException:
            logging.getLogger('cterasdk.edge').error("Login failed. %s", {'host': host, 'user': username})
            raise

    def sso(self, ticket):
        """
        Single Sign On

        :param str ticket: SSO Ticket.
        """
        logging.getLogger('cterasdk.edge').info("Performing Single Sign On.")
        self._edge.api.get('/ssologin', params={'ticket': ticket})
        self._edge.ctera_migrate.login()

    def logout(self):
        host = self._edge.host()
        user = self._edge.session().account.name
        try:
            self._edge.api.form_data('/logout', {'foo': 'bar'})
            logging.getLogger('cterasdk.edge').info("User logged out. %s", {'host': host, 'user': user})
        except CTERAException:
            logging.getLogger('cterasdk.edge').error("Logout failed. %s", {'host': host, 'user': user})
            raise



import cterasdk.settings
from ..services import CTERA, client_settings
from ..endpoints import EndpointBuilder
from ...clients.asynchronous import clients

from .. import authenticators

from ...lib.session.core import Session

from ...asynchronous.core import login
from ...asynchronous.core import cloudfs
from ...asynchronous.core import notifications
from ...asynchronous.core import users


class Clients:

    def __init__(self, core):
        self.v1 = V1(core)
        self.v2 = V2(core)
        self.io = IO(core)


class V1:

    def __init__(self, core):
        self.api = core.default.clone(clients.AsyncAPI, EndpointBuilder.new(core.base, core.context, '/api'))


class V2:

    def __init__(self, core):
        self.api = core.default.clone(clients.AsyncJSON, EndpointBuilder.new(core.base, core.context, '/v2/api'))


class IO:

    def __init__(self, core):
        self.webdav = WebDAV(core)


class WebDAV:

    def __init__(self, core):
        self._webdav = core.default.clone(clients.AsyncWebDAV, EndpointBuilder.new(core.base, core.context, '/webdav'))

    @property
    def download(self):
        return self._webdav.get


class AsyncPortal(CTERA):

    async def __aenter__(self):
        return self

    def __init__(self, host, port=None, https=True):
        super().__init__(host, port, https, base=None)
        self._default = clients.AsyncClient(EndpointBuilder.new(self.base),
                                            settings=client_settings(cterasdk.settings.sessions.metadata_connector),
                                            authenticator=self._authenticator)
        self._ctera_session = Session(self.host(), self.context)
        self._ctera_clients = Clients(self)

        self.cloudfs = cloudfs.CloudFS(self)
        self.notifications = notifications.Notifications(self)
        self.users = users.Users(self)

    @property
    def default(self):
        return self._default

    @property
    def v1(self):
        return self.clients.v1

    @property
    def v2(self):
        return self.clients.v2

    @property
    def io(self):
        return self.clients.io

    async def login(self, username, password):
        self._before_login()
        await self._login_object.login(username, password)
        await self._ctera_session.async_start_session(self)
        self._after_login()

    async def logout(self):
        if self._ctera_session.connected:
            await self._login_object.logout()
            self._ctera_session.stop_session()
        await self.default.close()

    @property
    def _login_object(self):
        return login.Login(self)

    def _authenticator(self, url):
        return authenticators.core(self.session(), url, self.context)

    async def __aexit__(self, exc_type, exc, tb):
        await self.default.close()


class AsyncGlobalAdmin(AsyncPortal):

    @property
    def context(self):
        return 'admin'


class AsyncServicesPortal(AsyncPortal):

    @property
    def context(self):
        return 'ServicesPortal'



import logging
import snappy
from .exceptions import DirectIOError


def decompress(compressed_block):
    """
    Decompress a Block.

    :param bytes block: Compressed Block
    :returns: Decompressed Block
    :rtype: bytes
    """
    return decompress_with_magic_header(compressed_block) \
        if compressed_block.startswith(b'\x82SNAPPY\x00') else decompress_without_magic_header(compressed_block)


def decompress_with_magic_header(compressed_block):
    """
    Decompress a Block.

    :param bytes block: Compressed Block
    :returns: Decompressed Block
    :rtype: bytes
    """
    try:
        logging.getLogger('cterasdk.direct').debug('Decompressing Block.')
        decompressed_block = bytearray()
        size_of_data = len(compressed_block)
        chunk_size, chunk_start = 4, 16
        while chunk_start < size_of_data:
            chunk_end = chunk_start + chunk_size + int.from_bytes(compressed_block[chunk_start:chunk_start + chunk_size])
            chunk_start = chunk_start + chunk_size
            if chunk_end > size_of_data:
                break
            decompressed_block.extend(snappy.decompress(compressed_block[chunk_start:chunk_end]))
            chunk_start = chunk_end
        return decompressed_block
    except snappy.UncompressError:
        raise DirectIOError()


def decompress_without_magic_header(compressed_block):
    """
    Decompress a Block.

    :param bytes compressed_block: Compressed Block
    :returns: Decompressed Block
    :rtype: bytes
    """
    return snappy.uncompress(compressed_block[:compressed_block.rfind(b'EOF') + 3])



import logging

from . import enum
from ..common import Object
from .base_command import BaseCommand


class Audit(BaseCommand):
    """
    Edge Filer Audit configuration APIs

    :ivar list[cterasdk.edge.enum.AuditEvents] defaultAuditEvents: Default audit events
    """

    defaultAuditEvents = [
        enum.AuditEvents.CreateFilesWriteData,
        enum.AuditEvents.CreateFoldersAppendData,
        enum.AuditEvents.WriteExtendedAttributes,
        enum.AuditEvents.DeleteSubfoldersAndFiles,
        enum.AuditEvents.WriteAttributes,
        enum.AuditEvents.Delete,
        enum.AuditEvents.ChangePermissions,
        enum.AuditEvents.ChangeOwner
    ]

    def enable(
            self,
            path,
            auditEvents=None,
            logKeepPeriod=30,
            maxLogKBSize=102400,
            maxRotateTime=1440,
            includeAuditLogTag=True,
            humanReadableAuditLog=False):
        """
        Enable Edge Filer Audit log

        :param str path: Path to save the audit log
        :param list[cterasdk.edge.enum.AuditEvents],optional auditEvents:
         List of audit event types to save, defaults to Audit.defaultAuditEvents
        :param int,optional logKeepPeriod: Period to key the logs in days, defaults to 30
        :param int,optional maxLogKBSize: The maximum size of the log file in KB, defailts to 102400 (100 MB)
        :param int,optional maxRotateTime: The maximal time before rotating the log file in Minutes, defaults to 1440 (24 hours)
        :param bool,optional includeAuditLogTag: Include audit log tag, defailts to True
        :param bool,optional humanReadableAuditLog: Human readable audit log, defailts to False
        """
        settings = Object()
        settings.mode = enum.Mode.Enabled
        settings.path = path
        settings.auditEvents = auditEvents or Audit.defaultAuditEvents
        settings.logKeepPeriod = logKeepPeriod
        settings.maxLogKBSize = maxLogKBSize
        settings.maxRotateTime = maxRotateTime
        settings.includeAuditLogTag = includeAuditLogTag
        settings.humanReadableAuditLog = humanReadableAuditLog

        logging.getLogger('cterasdk.edge').info('Enabling SMB audit logs')
        self._edge.api.put('/config/logging/files', settings)
        logging.getLogger('cterasdk.edge').info('Audit logs enabled')

    def disable(self):
        """
        Disable Edge Filer Audit log
        """
        logging.getLogger('cterasdk.edge').info('Disabling SMB audit logs')
        self._edge.api.put('/config/logging/files/mode', enum.Mode.Disabled)
        logging.getLogger('cterasdk.edge').info('Audit logs disabled')



import logging
import asyncio

from .types import DirectIOResponse, Block
from .crypto import decrypt_key, decrypt_block
from .decompressor import decompress
from .exceptions import UnAuthorized, UnprocessableContent, BlocksNotFoundError, DownloadError, DownloadTimeout, BlockListTimeout, \
    DownloadConnectionError, DecryptKeyError, DecryptBlockError, NotFoundError, DecompressBlockError, BlockValidationException, \
    BlockListConnectionError, DirectIOError

from ..exceptions import ClientResponseException


async def retry(coro, retries=3, backoff=1):
    """
    Retry Coroutine.

    :param coroutine coro: Asynchronous Coroutine.
    :param int retries: Retries.
    :param int backoff: Seconds.
    """
    attempts = 0
    while attempts < retries:
        try:
            return await coro()
        except DirectIOError as error:
            attempts = attempts + 1
            if attempts == retries:
                raise error
            wait = backoff * (2 ** (attempts - 1))
            logging.getLogger('cterasdk.direct').debug('Failed attempt number %s. Retrying in %s seconds.', attempts, wait)
            await asyncio.sleep(wait)


async def get_object(client, file_id, chunk):
    """
    Get Object from a Signed URL.

    :param int file_id: File ID.
    :param cterasdk.direct.types.Chunk chunk: Chunk.
    :returns: Object
    :rtype: bytes
    """
    async def get_object_coro():
        parameters = {'file_id': file_id, 'number': chunk.index, 'offset': chunk.offset}
        logging.getLogger('cterasdk.direct').debug('Downloading Block. %s', parameters)
        try:
            response = await client.get(chunk.location)
            return await response.read()
        except ConnectionError:
            logging.getLogger('cterasdk.direct').error('Failed to download block. Connection error. %s', parameters)
            raise DownloadConnectionError(file_id, chunk)
        except asyncio.TimeoutError:
            logging.getLogger('cterasdk.direct').error('Failed to download block. Timed out. %s', parameters)
            raise DownloadTimeout(file_id, chunk)
        except ClientResponseException as error:
            logging.getLogger('cterasdk.direct').error('Failed to download block. Error. %s', parameters)
            raise DownloadError(error.response, file_id, chunk)

    return await retry(get_object_coro)


async def decrypt_object(file_id, encrypted_object, encryption_key, chunk):
    """
    Decrypt Encrypted Object.

    :param bytes encrypted_object: Encrypted object.
    :param bytes encryption_key: Encryption key.
    :param cterasdk.direct.types.Chunk chunk: Chunk.
    :returns: Decrypted Object.
    :rtype: bytes
    """
    try:
        return decrypt_block(encrypted_object, encryption_key)
    except DirectIOError:
        logging.getLogger('cterasdk.direct').error('Failed to decrypt block.')
        raise DecryptBlockError(file_id, chunk)


async def decompress_object(file_id, compressed_object, chunk):
    """
    Decompress Object.

    :param bytes compressed_object: Compressed object.
    :param cterasdk.direct.types.Chunk chunk: Chunk.
    :returns: Decompressed Object.
    :rtype: bytes
    """
    try:
        decompressed_object = decompress(compressed_object)
        if chunk.length != len(decompressed_object):
            logging.getLogger('cterasdk.direct').error('Expected block length does not match decrypted and decompressed block length.')
            raise BlockValidationException(file_id, chunk)
        return decompressed_object
    except DirectIOError:
        logging.getLogger('cterasdk.direct').error('Failed to decompress block.')
        raise DecompressBlockError(file_id, chunk)


async def process_chunk(client, file_id, chunk, encryption_key, semaphore):
    """
    Process a Chunk.

    :param cterasdk.clients.asynchronous.clients.AsyncClient client: Asynchronous HTTP Client.
    :param int file_id: File ID.
    :param cterasdk.direct.types.Chunk chunk: Chunk.
    :param str encryption_key: Encryption key.
    :param asyncio.Semaphore semaphore: Semaphore.

    :returns: Block
    :rtype: cterasdk.direct.types.Block
    """
    async def process(client, chunk, encryption_key):
        parameters = {'file_id': file_id, 'number': chunk.index, 'offset': chunk.offset}
        logging.getLogger('cterasdk.direct').debug('Processing Block. %s', parameters)

        encrypted_object = await get_object(client, file_id, chunk)
        decrypted_object = await decrypt_object(file_id, encrypted_object, encryption_key, chunk)
        decompressed_object = await decompress_object(file_id, decrypted_object, chunk)
        return Block(file_id, chunk.index, chunk.offset, decompressed_object, chunk.length)

    if semaphore is not None:
        async with semaphore:
            return await process(client, chunk, encryption_key)
    return await process(client, chunk, encryption_key)


async def process_chunks(client, file_id, chunks, encryption_key, semaphore=None):
    """
    Process Chunks Asynchronously.

    :param cterasdk.clients.asynchronous.clients.AsyncClient client: Asynchronous HTTP Client.
    :param int file_id: File ID.
    :param list[cterasdk.direct.types.Chunk] chunks: Chunk.
    :param str encryption_key: Encryption key.
    :param asyncio.Semaphore,optional semaphore: Semaphore.
    :returns: List of futures.
    :rtype: list[asyncio.Task]
    """
    parameters = {'file_id': file_id, 'blocks': len(chunks)}
    if semaphore:
        parameters['max_workers'] = semaphore._value  # pylint: disable=protected-access
    logging.getLogger('cterasdk.direct').debug('Processing Blocks. %s', parameters)
    futures = []
    for chunk in chunks:
        futures.append(asyncio.create_task(process_chunk(client, file_id, chunk, encryption_key, semaphore)))
    return futures


def decrypt_encryption_key(file_id, wrapped_key, secret_access_key):
    """
    Decrypt Encryption Key.

    :param int file_id: File ID.
    :param str wrapped_key: Encryption Key.
    :param str secret_access_key: Secret Access Key.
    :returns: Decrypted Encryption Key.
    :rtype: bytes
    """
    try:
        return decrypt_key(wrapped_key, secret_access_key)
    except DirectIOError:
        logging.getLogger('cterasdk.direct').error('Failed to decrypt secret key.')
        raise DecryptKeyError(file_id)


def create_authorization_header(credentials):
    """
    Create Authorization Header.

    :param cterasdk.objects.asynchronous.direct.Credentials credentials: Credentials
    :returns: Authorization header as a dictionary.
    :rtype: dict
    """
    return {
        'Authorization': f'Bearer {credentials.access_key_id}'
    }


async def get_chunks(api, credentials, file_id):
    """
    Get Chunks.

    :param cterasdk.clients.asynchronous.clients.AsyncJSON api: Asynchronous JSON Client.
    :param int file_id: File ID.
    :returns: Wrapped key and file chunks.
    :rtype: cterasdk.direct.types.DirectIOResponse
    """
    async def get_chunks_coro():
        parameters = {'file_id': file_id}
        logging.getLogger('cterasdk.direct').debug('Listing blocks. %s', parameters)
        try:
            response = await api.get(f'{file_id}', headers=create_authorization_header(credentials))
            if not response.chunks:
                logging.getLogger('cterasdk.direct').error('Blocks not found. %s', parameters)
                raise BlocksNotFoundError(file_id)
            return DirectIOResponse(response)
        except ClientResponseException as error:
            if error.response.status == 400:
                raise NotFoundError(file_id)
            if error.response.status == 401:
                raise UnAuthorized(file_id)
            if error.response.status == 422:
                raise UnprocessableContent(file_id)
            raise error
        except ConnectionError:
            logging.getLogger('cterasdk.direct').error('Failed to list blocks. Connection error. %s', parameters)
            raise BlockListConnectionError(file_id)
        except asyncio.TimeoutError:
            logging.getLogger('cterasdk.direct').error('Failed to list blocks. Timed out. %s', parameters)
            raise BlockListTimeout(file_id)

    return await retry(get_chunks_coro)



import urllib.parse
from pathlib import Path


def create(base, *segments):
    """
    Create a URL from a Base URL, and URL segments

    :param str base: Base URL
    :param list(str),optional segments: URL segments

    :returns: Full URL, comprised of the Base URL and URL segments
    :rtype: str
    """
    url_components = components(base)
    segments = [url_components.path, *segments]
    root = '/'
    url = Path(root)
    for segment in segments:
        segment = segment[1:] if Path(segment).is_relative_to(root) else segment
        url = url.joinpath(segment)
    url = urllib.parse.urljoin(base, url.as_posix())
    return f'{url}{root}' if segments[-1] == root else url


def join(base, url):
    return urllib.parse.urljoin(base, url)


def components(url):
    """
    Parse a URL
    """
    return urllib.parse.urlparse(url)


def unquote(url):
    return urllib.parse.unquote(url)


def quote(url):
    return urllib.parse.quote(url)



import logging
from ..exceptions import CTERAException
from ..common.utils import tcp_connect


def test(Edge):
    tcp_connect(Edge.host(), Edge.port())
    logging.getLogger('cterasdk.edge').debug('Trying to obtain login info.')
    response = Edge.api.get('/nosession/logininfo')
    if response is not None:
        logging.getLogger('cterasdk.edge').debug('Successfully obtained login info. %s', {'hostname': response.hostname})
        return response
    raise CTERAException('Could not obtain login info for device {host}:{port}.')



import asyncio
from collections import namedtuple

import cterasdk.settings

from . import filters
from .lib import get_chunks, decrypt_encryption_key, process_chunks
from .types import File, ByteRange, FileMetadata
from .stream import Streamer

from ..objects.endpoints import DefaultBuilder, EndpointBuilder
from ..clients.settings import ClientSettings, ClientTimeout, TCPConnector
from ..clients.asynchronous.clients import AsyncClient, AsyncJSON


Credentials = namedtuple('Credentials', ('access_key_id', 'secret_access_key'))
Credentials.__doc__ = 'Tuple holding the access and secret keys to access objects using DirectIO'
Credentials.access_key_id.__doc__ = 'Access key'
Credentials.secret_access_key.__doc__ = 'Secret Key'


def client_settings(parameters):
    return ClientSettings(
        TCPConnector(parameters.ssl),
        ClientTimeout(**parameters.timeout.kwargs)
    )


class Client:

    def __init__(self, baseurl, credentials):
        """
        :param str baseurl: Portal URL
        :param cterasdk.objects.asynchronous.directio.Credentials credentials: Credentials
        """
        self._api = AsyncJSON(EndpointBuilder.new(baseurl, '/directio'),
                              settings=client_settings(cterasdk.settings.sessions.ctera_direct.api),
                              authenticator=lambda *_: True)
        self._client = AsyncClient(DefaultBuilder(),
                                   settings=client_settings(cterasdk.settings.sessions.ctera_direct.storage),
                                   authenticator=lambda *_: True)
        self._credentials = credentials

    async def _direct(self, file_id):
        server_object = await get_chunks(self._api, self._credentials, file_id)
        encryption_key = decrypt_encryption_key(file_id, server_object.wrapped_key, self._credentials.secret_access_key)
        return File(file_id, encryption_key, server_object.chunks)

    async def metadata(self, file_id):
        """
        Direct IO Metadata API.

        :param int file_id: File ID.
        """
        return FileMetadata(await self._direct(file_id))

    async def blocks(self, file_id, blocks, max_workers):
        """
        Blocks API.

        :param int file_id: File ID.
        :param list[cterasdk.direct.exceptions.BlockInfo] blocks: List of BlockInfo objects,
         or list of integers identifying the block position.
        :param int max_workers: Max concurrent tasks. A task will be dispatched for each block if no limited was specified.
        :returns: List of Blocks.
        :rtype: list[cterasdk.direct.types.Block]
        """
        file = await self._direct(file_id)
        executor = await self.executor(filters.blocks(file, blocks), file.encryption_key, file_id, max_workers)
        return await executor()

    async def streamer(self, file_id, byte_range):
        """
        Stream API.

        :param int file_id: File ID.
        :param cterasdk.direct.types.ByteRange byte_range: Byte Range.
        :returns: Streamer Object
        :rtype: cterasdk.direct.stream.Streamer
        """
        file = await self._direct(file_id)
        byte_range = byte_range if byte_range is not None else ByteRange.default()
        max_workers = cterasdk.settings.sessions.ctera_direct.streamer.max_workers
        executor = await self.executor(filters.span(file, byte_range), file.encryption_key, file_id, max_workers)
        return Streamer(executor, byte_range)

    async def executor(self, chunks, encryption_key, file_id=None, max_workers=None):
        """
        Get Blocks.

        :param list[cterasdk.direct.types.Chunk] chunks: List of Chunks.
        :param str encryption_key: Decryption Key.
        :param int,optional file_id: File ID.
        :param int,optional max_workers: Max concurrent tasks.

        :returns: Callable Downloader
        :rtype: function
        """

        async def execute():
            """
            Asynchronous Executable of Chunk Retrieval Tasks.
            """
            return await process_chunks(self._client, file_id, chunks, encryption_key,
                                        asyncio.Semaphore(max_workers) if max_workers else None)

        return execute

    async def close(self):
        await self._api.close()
        await self._client.close()


class DirectIO:

    async def __aenter__(self):
        return self

    def __init__(self, baseurl=None, access_key_id=None, secret_access_key=None):
        """
        Initialize a DirectIO Client.

        :param str baseurl: Portal URL
        :param str,optional access_key_id: Access key
        :param str,optional secret_access_key: Secret key
        """
        self._client = Client(baseurl, Credentials(access_key_id, secret_access_key))

    async def metadata(self, file_id):
        """
        Get Metadata.

        :param int file_id: File ID
        """
        return await self._client.metadata(file_id)

    async def blocks(self, file_id, blocks=None, max_workers=None):
        """
        Get Blocks.

        :param int file_id: File ID
        :param list[cterasdk.direct.exceptions.BlockInfo] blocks: List of BlockInfo objects,
         or list of integers identifying the block position.
        :param int max_workers: Max concurrent tasks. A task will be dispatched for each block if no limited was specified.
        :returns: Blocks
        :rtype: list[cterasdk.direct.types.Block]
        """
        return await self._client.blocks(file_id, blocks, max_workers)

    async def streamer(self, file_id, byte_range=None):
        """
        Iterates over data chunks.

        :param int file_id: File ID.
        :param cterasdk.direct.types.ByteRange,optional byte_range: Byte Range.
        :returns: Stream Object
        :rtype: cterasdk.direct.stream.Streamer
        """
        return await self._client.streamer(file_id, byte_range)

    async def close(self):
        await self._client.close()

    async def __aexit__(self, exc_type, exc, tb):
        await self.close()



import time
import logging
from abc import ABC, abstractmethod
from ..exceptions import CTERAException
from ..convert import tojsonstr


class TaskRunningStatus:
    Running = 'running'
    Completed = 'completed'
    Failed = 'failed'
    Warnings = 'completed with warnings'


class TaskError(CTERAException):

    def __init__(self, task):
        super().__init__()
        self.task = task


class TaskBase(ABC):

    def __init__(self, CTERAHost, ref, retries=10, seconds=1):
        self.CTERAHost = CTERAHost
        self.path = self._get_task_id(ref)
        self.attempt = 0
        self.retries = retries
        self.seconds = seconds
        self.running = True

    @abstractmethod
    def _get_task_id(self, ref):
        raise NotImplementedError("Subclass must implement _get_task_id")

    @abstractmethod
    def get_task_status(self):
        raise NotImplementedError("Subclass must implement get_task_status")

    def wait(self):
        task = None
        while self.running:
            logging.getLogger('cterasdk.common').debug('Obtaining task status. %s', {'path': self.path, 'attempt': (self.attempt + 1)})
            task = self.get_task_status()
            logging.getLogger('cterasdk.common').debug('Task status. %s', tojsonstr(task, False))
            self.increment()
            self.running = task.status == TaskRunningStatus.Running
        return TaskBase.resolve(task)

    @staticmethod
    def resolve(task):
        task_info = {'id': task.id, 'name': task.name, 'status': task.status, 'start_time': task.startTime, 'end_time': task.endTime}
        if task.status == TaskRunningStatus.Failed:
            logging.getLogger('cterasdk.common').error('Task failed. %s', task_info)
            raise TaskError(task)
        if task.status == TaskRunningStatus.Warnings:
            logging.getLogger('cterasdk.common').warning('Task completed with warnings. %s', task_info)
        if task.status == TaskRunningStatus.Completed:
            logging.getLogger('cterasdk.common').debug('Task completed successfully. %s', task_info)
        return task

    def increment(self):
        if self.attempt >= self.retries:
            duration = time.strftime("%H:%M:%S", time.gmtime(self.retries * self.seconds))
            logging.getLogger('cterasdk.common').error('Could not obtain task status in a timely manner. %s', {'duration': duration})
            raise CTERAException('Timed out. Could not obtain task status in a timely manner', None, duration=duration)
        self.attempt = self.attempt + 1
        logging.getLogger('cterasdk.common').debug('Sleep. %s', {'seconds': self.seconds})
        time.sleep(self.seconds)



from . import uri


def edge(ctera_session, url):
    if ctera_session.connecting or \
       ctera_session.connected or \
       uri.components(url).path.endswith(_edge_no_session_resources()):
        return True
    return False


def core(ctera_session, url, context):
    if ctera_session.connecting or \
       ctera_session.connected or \
       uri.components(url).path.startswith(_core_no_session_resources(context)):
        return True
    return False


def _edge_no_session_resources():
    return _no_session_resources('', '/admingui/api/login', '/ssologin', '/api/nosession/logininfo',
                                 '/api/nosession/createfirstuser', '/migration/rest/v1/auth/user')


def _core_no_session_resources(context):
    return _no_session_resources(context, '/api/login', '/public', '/setup', '/startup')


def _no_session_resources(context, *paths):
    return tuple(uri.create(context, path) for path in paths)



import logging
import functools

from ..exceptions import CTERAException


def authenticated(execute_request):
    @functools.wraps(execute_request)
    def authenticate_then_execute(self, *args, **kwargs):
        if callable(self._authenticator) and self._authenticator(self._builder(args[0])):  # pylint: disable=protected-access
            return execute_request(self, *args, **kwargs)
        logging.getLogger('cterasdk.common').error('Not logged in.')
        raise CTERAException('Not logged in')
    return authenticate_then_execute



import logging

from .directorytree import DirectoryTree

from .enum import OperationMode
from .base_command import BaseCommand


class Cache(BaseCommand):
    """ Edge Filer cache configuration """

    def enable(self):
        """ Enable caching """
        logging.getLogger('cterasdk.edge').info('Enabling caching.')
        self._set_operation_mode(OperationMode.CachingGateway)

    def disable(self):
        """ Disable caching """
        logging.getLogger('cterasdk.edge').info('Disabling caching.')
        self._set_operation_mode(OperationMode.Disabled)

    def force_eviction(self):
        """ Force eviction """
        logging.getLogger('cterasdk.edge').info("Starting file eviction.")
        self._edge.api.execute("/config/cloudsync", "forceExecuteEvictor", None)
        logging.getLogger('cterasdk.edge').info("Eviction started.")

    def is_enabled(self):
        return self._edge.api.get('/config/cloudsync/cloudExtender/operationMode') == OperationMode.CachingGateway

    def _set_operation_mode(self, mode):
        self._edge.api.put('/config/cloudsync/cloudExtender/operationMode', mode)
        logging.getLogger('cterasdk.edge').info('Device opreation mode changed. %s', {'mode': mode})

    def pin(self, path):
        """
        Pin a folder

        :param str path: Directory path
        """
        directory_tree = self._fetch_pinning_config()
        logging.getLogger('cterasdk.edge').info('Pinning folder. %s', {'path': path})
        directory_tree.include_folder(path)
        self._update_pinning_config(directory_tree)

    def pin_exclude(self, path):
        """
        Exclude a sub-folder from a pinned folder

        :param str path: Directory path
        """
        directory_tree = self._fetch_pinning_config()
        logging.getLogger('cterasdk.edge').info('Excluding sub-folder. %s', {'path': path})
        directory_tree.exclude_folder(path)
        self._update_pinning_config(directory_tree)

    def remove_pin(self, path):
        """
        Remove a pin from a previously pinned folder

        :param str path: Directory path
        """
        directory_tree = self._fetch_pinning_config()
        logging.getLogger('cterasdk.edge').info('Removing pin from previously pinned folder. %s', {'path': path})
        directory_tree.remove_selection(path)
        self._update_pinning_config(directory_tree)

    def pin_all(self):
        """ Pin all folders """
        directory_tree = self._fetch_pinning_config()
        logging.getLogger('cterasdk.edge').info('Pinning all folders.')
        directory_tree.select_all()
        self._update_pinning_config(directory_tree)

    def unpin_all(self):
        """ Remove all folder pins """
        directory_tree = self._fetch_pinning_config()
        logging.getLogger('cterasdk.edge').info('Removing all folder pins.')
        directory_tree.unselect_all()
        self._update_pinning_config(directory_tree)

    def _fetch_pinning_config(self):
        root = self._edge.api.get('/config/cloudsync/cloudExtender/selectedFolders')
        return DirectoryTree(root)

    def _update_pinning_config(self, directory_tree):
        self._edge.api.put('/config/cloudsync/cloudExtender/selectedFolders', directory_tree.root)

    def pin_recursive(self, path):
        """
        Pin a folder and all its subfolders recursively

        :param str path: Directory path
        """
        directory_tree = self._fetch_pinning_config()
        logging.getLogger('cterasdk.edge').info('Recursively pinning folder and all subfolders. %s', {'path': path})
        directory_tree.include_folder_recursive(path)
        self._update_pinning_config(directory_tree)

    def unpin_recursive(self, path):
        """
        Unpin a folder and all its subfolders recursively

        :param str path: Directory path
        """
        directory_tree = self._fetch_pinning_config()
        logging.getLogger('cterasdk.edge').info('Recursively unpinning folder and all subfolders. %s', {'path': path})
        directory_tree.exclude_folder_recursive(path)
        self._update_pinning_config(directory_tree)



class Mode:
    """
    Enum for operational mode

    :ivar str Enabled: Operational mode enabled
    :ivar str Disabled: Operational mode diabled
    """
    Enabled = "enabled"
    Disabled = "disabled"


class IPProtocol:
    """
    IP Protocol

    :ivar str TCP: TCP Protocol
    :ivar str UDP: UDP Protocol
    """
    TCP = "TCP"
    UDP = "UDP"


class Severity:
    """
    Log severity levels

    :ivar str EMERGENCY: Emergency log level
    :ivar str ALERT: Alert log level
    :ivar str CRITICAL: Critical log level
    :ivar str ERROR: Error log level
    :ivar str WARNING: Warning log level
    :ivar str NOTICE: Notice log level
    :ivar str INFO: Info log level
    :ivar str DEBUG: Debug log level
    """
    EMERGENCY = "emergency"
    ALERT = "alert"
    CRITICAL = "critical"
    ERROR = "error"
    WARNING = "warning"
    NOTICE = "notice"
    INFO = "info"
    DEBUG = "debug"


class VolumeStatus:
    """
    Edge Filer volume status

    :ivar str Ok: Volume is ok
    :ivar str ContainsErrors: Volume contains errors
    :ivar str ReadOnly: Voilume is read only
    :ivar str Corrupted: Volume is corrupted"
    :ivar str Unknown: Volume status is unknown
    :ivar str Recovering: Volume is recovering
    :ivar str Mounting: Volume is mounting
    :ivar str Unmounted: Volume is unmounted
    :ivar str Formatting: Volume is formatting
    :ivar str Converting: Volume is converting
    :ivar str Resizing: Volume is resizing
    :ivar str Repairing: Volume is repairing
    :ivar str Checking: Volume is checking
    :ivar str KeyRequired: Volume required key
    :ivar str CheckingQuota: Checking volume  quota
    """
    Ok = "ok"
    ContainsErrors = "containsErrors"
    ReadOnly = "readOnly"
    Corrupted = "corrupted"
    Unknown = "unknown"
    Recovering = "recovering"
    Mounting = "mounting"
    Unmounted = "unmounted"
    Formatting = "formatting"
    Converting = "converting"
    Resizing = "resizing"
    Repairing = "repairing"
    Checking = "checking"
    KeyRequired = "keyRequired"
    CheckingQuota = "checkingQuota"


class CIFSPacketSigning:
    """
    CIFS Packet signing options

    :ivar str Disabled: CIFS Packet signing is disabled
    :ivar str IfClientAgrees: Use CIFS Packet signing is client agrees
    :ivar str Required: Require CIFS Packet signing
    """
    Disabled = "Disabled"
    IfClientAgrees = "If client agrees"
    Required = "Required"


class TaskStatus:
    """
    Edge Filer task status

    :ivar str Failed: The task has failed
    :ivar str Running: The task is running
    :ivar str Completed: The task has completed
    """
    Failed = "failed"
    Running = "running"
    Completed = "completed"


class TCPConnectRC:
    Open = "Open"


class License:
    """
    Edge Filer license types

    :ivar str EV4: EV4 license
    :ivar str EV8: EV8 license
    :ivar str EV16: EV16 license
    :ivar str EV32: EV32 license
    :ivar str EV64: EV64 license
    :ivar str EV128: EV128 license
    """
    EV4 = "EV4"
    EV8 = "EV8"
    EV16 = "EV16"
    EV32 = "EV32"
    EV64 = "EV64"
    EV128 = "EV128"


class OperationMode:
    """
    Edge Filer operation mode

    :ivar str Disabled: Edge Filer is Disabled
    :ivar str CachingGateway: Edge Filer is in Caching mode
    """
    Disabled = "Disabled"
    CachingGateway = "CachingGateway"


class Acl:
    """
    ACL types

    :ivar str WindowsNT: Windows NT ACL Mode
    :ivar str OnlyAuthenticatedUsers: Authenticated Users ACL Mode
    """
    WindowsNT = "winAclMode"
    OnlyAuthenticatedUsers = "authenticated"


class ClientSideCaching:
    """
    Client side caching types

    :ivar str Manual: Manual client side caching
    :ivar str Documents: Documents client side caching
    :ivar str Disabled: Client side caching disabled
    """
    Manual = "manual"
    Documents = "documents"
    Disabled = "disable"


class PrincipalType:
    """
    ACL Principal Type

    :ivar str LU: Local User
    :ivar str LG: Local Group
    :ivar str DU: Domain User
    :ivar str DG: Domain Group
    """
    LU = "LocalUser"
    LG = "LocalGroup"
    DU = "DomainUser"
    DG = "DomainGroup"


class LocalGroup:
    """
    Local Group types

    :ivar str Administrators: Administrators
    :ivar str ReadOnlyAdministrators: Read Only Administrators
    :ivar str Everyone: Everyone

    """
    Administrators = "Administrators"
    ReadOnlyAdministrators = "Read Only Administrators"
    Everyone = "Everyone"


class FileAccessMode:
    """
    File Access Mode

    :ivar str RW: Read Write
    :ivar str RO: Read Only
    :ivar str NA: None
    """
    RW = "ReadWrite"
    RO = "ReadOnly"
    NA = "None"


class SyncStatus:
    """
    Edge Filer sync status

    :ivar str Off: Off
    :ivar str NotInitialized: Not Initialized
    :ivar str InitializingConnection: Initializing Connection
    :ivar str ConnectingFolders: Connecting Folders
    :ivar str Connected: Connected
    :ivar str ClocksOutOfSync: Clocks is Out Of Sync
    :ivar str ConnectionFailed: Connection Failed
    :ivar str InternalError: Internal Error
    :ivar str InvalidConfiguration: Invalid Configuration
    :ivar str VolumeUnavailable: Volume Unavailable
    :ivar str NoFolder: No Folder
    :ivar str DisconnectedPortal: Disconnected from Portal
    :ivar str ServiceUnavailable: Service is Unavailable
    :ivar str Unlicensed: Unlicensed
    :ivar str Synced: Synced
    :ivar str Syncing: Syncing
    :ivar str Scanning: Scanning
    :ivar str UpgradingDataBase: Upgrading Database
    :ivar str OutOfQuota: Out of Quota
    :ivar str RejectedByPolicy: Rejected by Policy
    :ivar str FailedFilesInReadOnlyFolder: Failed Files in Read Only Folder
    :ivar str ShouldSupportWinNtAcl: Should Support WinNt Acl
    :ivar str TakingSnapshot: Taking Snapshot
    :ivar str CatalogReadOnlyMode: Catalog ReadOnly Mode
    :ivar str InvalidAverageBlockSize: Invalid Average Block Size
    """
    Off = "Off"
    NotInitialized = "NotInitialized"
    InitializingConnection = "InitializingConnection"
    ConnectingFolders = "ConnectingFolders"
    Connected = "Connected"
    ClocksOutOfSync = "ClocksOutOfSync"
    ConnectionFailed = "ConnectionFailed"
    InternalError = "InternalError"
    InvalidConfiguration = "InvalidConfiguration"
    VolumeUnavailable = "VolumeUnavailable"
    NoFolder = "NoFolder"
    DisconnectedPortal = "DisconnectedPortal"
    ServiceUnavailable = "ServiceUnavailable"
    Unlicensed = "Unlicensed"
    Synced = "Synced"
    Syncing = "Syncing"
    Scanning = "Scanning"
    UpgradingDataBase = "UpgradingDataBase"
    OutOfQuota = "OutOfQuota"
    RejectedByPolicy = "RejectedByPolicy"
    FailedFilesInReadOnlyFolder = "FailedFilesInReadOnlyFolder"
    ShouldSupportWinNtAcl = "ShouldSupportWinNtAcl"
    TakingSnapshot = "TakingSnapshot"
    CatalogReadOnlyMode = "CatalogReadOnlyMode"
    InvalidAverageBlockSize = "InvalidAverageBlockSize"


class AuditEvents:
    """
    Audit log event types

    :ivar str ListFolderReadData: List Folder Read Data
    :ivar str CreateFilesWriteData: Create Files Write Data
    :ivar str CreateFoldersAppendData: Create Folders Append Data
    :ivar str ReadExtendedAttributes: Read Extended Attributes
    :ivar str WriteExtendedAttributes: Write Extended Attributes
    :ivar str TraverseFolderExecuteFile: Traverse Folder Execute File
    :ivar str DeleteSubfoldersAndFiles: Delete Subfolders And Files
    :ivar str WriteAttributes: Write Attributes
    :ivar str Delete: Delete
    :ivar str ChangePermissions: Change Permissions
    :ivar str ChangeOwner: Change Owner
    """
    ListFolderReadData = "RD"
    CreateFilesWriteData = "WD"
    CreateFoldersAppendData = "AD"
    ReadExtendedAttributes = "REA"
    WriteExtendedAttributes = "WEA"
    TraverseFolderExecuteFile = "X"
    DeleteSubfoldersAndFiles = "DC"
    WriteAttributes = "WA"
    Delete = "DE"
    ChangePermissions = "WDAC"
    ChangeOwner = "WO"


class ServicesConnectionState:
    """
    Edge Filer connection status

    :ivar str ResolvingServers: The Edge Filer is resolving CTERA Portal servers
    :ivar str Connecting: The Edge Filer is in connecting to CTERA Portal
    :ivar str Attaching: The Edge Filer is attaching to CTERA Portal
    :ivar str Authenticating: The Edge Filer is authenticating to CTERA Portal
    :ivar str Disconnected: The Edge Filer is disconnected from CTERA Portal
    :ivar str Connected: The Edge Filer is connected to CTERA Portal
    """
    ResolvingServers = "ResolvingServers"
    Connecting = "Connecting"
    Attaching = "Attaching"
    Authenticating = "Authenticating"
    Disconnected = "Disconnected"
    Connected = "Connected"


class RAIDLevel:
    """
    RAID Levels

    :ivar str JBOD: Linear concatenation
    :ivar str RAID_0: Stripe set
    :ivar str RAID_1: Mirror
    :ivar str RAID_5: Distributed parity
    :ivar str RAID_6: Dual parity
    """
    JBOD = "linear"
    RAID_0 = "0"
    RAID_1 = "1"
    RAID_5 = "5"
    RAID_6 = "6"
    LVM = 'LVM'


class BackupConfStatusID:
    """
    Status of backup configuration

    :ivar str NotInitialized: Backup configuration was not initialized
    :ivar str Configuring: Backup is being configured
    :ivar str Attaching: Backup configuration is Attaching
    :ivar str Attached: Backup configuration is Attached
    :ivar str NoFolder: No Folder for backup
    :ivar str WrongPassword: Wrong password used
    :ivar str Failed: Backup configuration failed
    :ivar str Unsubscribed: Unsubscribed to backup
    :ivar str Unlicensed: Unlicensed" to backup
    :ivar str ClocksOutOfSync: Clocks are out of sync
    :ivar str GetFoldersList: Get folders list
    """
    NotInitialized = "NotInitialized"
    Configuring = "Configuring"
    Attaching = "Attaching"
    Attached = "Attached"
    NoFolder = "NoFolder"
    WrongPassword = "WrongPassword"
    Failed = "Failed"
    Unsubscribed = "Unsubscribed"
    Unlicensed = "Unlicensed"
    ClocksOutOfSync = "ClocksOutOfSync"
    GetFoldersList = "GetFoldersList"


class Traffic:
    """
    Traffic type

    :ivar str Upload: Upload
    :ivar str Download: Download
    """
    Upload = 'Upload'
    Download = 'Download'


class SMBProtocol:
    """
    SMB Protocol

    :ivar str SMB1: SMB v1
    :ivar str NT1: SMB v1
    :ivar str SMB2_02: Vista, Server 2008
    :ivar str SMB2_10: Windows 7, Server 2008 R2
    :ivar str SMB3_00: Windows 8, Server 2012
    :ivar str SMB3_02: Windows 8.1, Server 2012 R2
    :ivar str SMB3_11: Windows 10, Server 2016
    :ivar str SMB2: Windows 7, Server 2008
    :ivar str SMB3: SMB 3.1.1
    """
    SMB1 = 'NT1'
    NT1 = SMB1
    SMB2_02 = 'SMB2_02'
    SMB2_10 = 'SMB2_10'
    SMB3_00 = 'SMB3_00'
    SMB3_02 = 'SMB3_02'
    SMB3_11 = 'SMB3_11'
    SMB2 = 'SMB2'
    SMB3 = 'SMB3'


class SourceType:
    """
    Source Host Type

    :ivar str Edge: This Edge Filer
    :ivar str Windows: Windows Server
    :ivar str ONTAP: NetApp ONTAP
    :ivar str OneFS: Isilon OneFS
    :ivar str Panzura: Panzura Freedom Filer
    :ivar str SGRID9_SMB: NetApp StorageGRID 9
    :ivar str SGRID11_SMB: NetApp StorageGRID 11
    :ivar str StorSimple: Microsoft Azure StorSimple
    """
    Edge = 'currentDevice'
    Windows = 'windowsServer'
    ONTAP = 'netapp'
    OneFS = 'isilon'
    Panzura = 'panzura'
    SGRID9_SMB = 'storageGrid9'
    SGRID11_SMB = 'storageGrid11'
    StorSimple = 'azureStorSimple'


class TaskType:
    """
    Migration Tool Task Type

    :ivar str Discovery: Discovery
    :ivar str Migration: Migration
    """
    Discovery = 0
    Migration = 1



import logging

from ..common import Object
from ..exceptions import CTERAException
from .base_command import BaseCommand


class StorageClasses(BaseCommand):
    """ Portal Storage Classes APIs """

    def add(self, name):
        """
        Add a storage class

        :param str name: Name
        """
        logging.getLogger('cterasdk.core').info("Adding storage class. %s", {'name': name})
        param = Object()
        param.name = name
        response = self._core.api.add('/storageClasses', param)
        logging.getLogger('cterasdk.core').info("Storage class added. %s", {'name': name})
        return response

    def all(self):
        """
        Get storage classes

        :returns: List of storage classes
        :rtype: list(cterasdk.common.object.Object)
        """
        if self._core.session().in_tenant_context():
            return self._core.api.execute('', 'getStorageClasses')
        return self._core.api.get('/storageClasses')

    def get(self, name):
        """
        Get storage class

        :param str name: Storage class name, defaults to ``None``
        :returns: Storage class
        :rtype: cterasdk.common.object.Object
        """
        if not self._core.session().in_tenant_context():
            return self._core.api.get(f'/storageClasses/{name}')
        for storage_class in self.all():
            if storage_class.name == name:
                return storage_class
        raise CTERAException('Could not find storage class.', name=name)



from ..base_command import BaseCommand
from . import io, common, file_access


class FileBrowser(BaseCommand):
    """ Edge Filer File Browser APIs """

    def __init__(self, edge):
        super().__init__(edge)
        self._file_access = file_access.FileAccess(edge)

    def listdir(self, path):
        """
        List Directory

        :param str path: Path
        """
        return io.listdir(self._edge, path)

    def download(self, path, destination=None):
        """
        Download a file

        :param str path: The file path on the Edge Filer
        :param str,optional destination:
         File destination, if it is a directory, the original filename will be kept, defaults to the default directory
        """
        return self._file_access.download(self.get_object_path(path), destination=destination)

    def download_as_zip(self, cloud_directory, files, destination=None):
        """
        Download a list of files and/or directories from a cloud folder as a ZIP file

        .. warning:: The list of files is not validated. The ZIP file will include only the existing  files and directories

        :param str cloud_directory: Path to the cloud directory
        :param list[str] files: List of files and/or directories in the cloud folder to download
        :param str,optional destination:
         File destination, if it is a directory, the filename will be calculated, defaults to the default directory
        """
        return self._file_access.download_as_zip(self.get_object_path(cloud_directory), files, destination=destination)

    def upload(self, path, destination):
        """
        Upload a file

        :param str path: Local path
        :param str destination: Remote path
        """
        self._file_access.upload(path, self.get_object_path(destination))

    def mkdir(self, path):
        """
        Create a new directory

        :param str path: Directory path
        """
        return io.mkdir(self._edge, self.get_object_path(path))

    def makedirs(self, path):
        """
        Create a directory recursively

        :param str path: Directory path
        """
        return io.makedirs(self._edge, self.get_object_path(path))

    def copy(self, path, destination=None, overwrite=False):
        """
        Copy a file or a folder

        :param str path: Source file or folder path
        :param str destination: Destination folder path
        :param bool,optional overwrite: Overwrite on conflict, defaults to False
        """
        if destination is None:
            raise ValueError('Copy destination was not specified.')
        return io.copy(self._edge, self.get_object_path(path), self.get_object_path(destination), overwrite)

    def move(self, path, destination=None, overwrite=False):
        """
        Move a file or a folder

        :param str path: Source file or folder path
        :param str destination: Destination folder path
        :param bool,optional overwrite: Overwrite on conflict, defaults to False
        """
        if destination is None:
            raise ValueError('Move destination was not specified.')
        return io.move(self._edge, self.get_object_path(path), self.get_object_path(destination), overwrite)

    def delete(self, path):
        """
        Delete a file

        :param str path: File path
        """
        return io.remove(self._edge, self.get_object_path(path))

    @staticmethod
    def get_object_path(path):
        return common.get_object_path('/', path)



from ...common import Object
from ...lib import FileAccessBase


class FileAccess(FileAccessBase):

    def _get_single_file_url(self, path):
        return path.fullpath()

    def _get_multi_file_url(self, cloud_directory, files):
        return '/admingui/api/status/fileManager/zip'

    @property
    def _use_file_url_for_multi_file_url(self):
        return False

    def _get_multi_file_object(self, cloud_directory, files):
        files_obj = Object()
        files_obj.paths = ['/'.join([cloud_directory.fullpath(), filename]) for filename in files]
        files_obj.snapshot = Object()
        files_obj._classname = 'BackupRepository'  # pylint: disable=protected-access
        files_obj.snapshot.location = 1
        files_obj.snapshot.timestamp = None
        files_obj.snapshot.path = None
        return files_obj

    def _get_upload_url(self, dest_path):
        return '/actions/upload'

    def _get_upload_form(self, local_file_info, fd, dest_path):
        return dict(
            name=local_file_info['name'],
            fullpath=f'{dest_path.fullpath()}/{local_file_info["name"]}',
            filedata=fd
        )

    def _get_zip_file_handle(self, cloud_directory, files):
        return self._ctera_host.io.download_zip(
            self._get_multi_file_url(cloud_directory, files),
            self._make_form_data(cloud_directory, files)
        )

    def _upload_object(self, local_file_info, fd, dest_path):
        return self._ctera_host.io.upload(
                self._get_upload_url(dest_path),
                self._get_upload_form(local_file_info, fd, dest_path)
            )



import base64
import logging
import binascii
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.exceptions import UnsupportedAlgorithm
from ..common.utils import utf8_decode
from .exceptions import DirectIOError


def decrypt_key(wrapped_key, secret):
    try:
        logging.getLogger('cterasdk.direct').debug('Decoding Secret.')
        decoded_secret = base64.b64decode(secret)
        decoded_secret = decoded_secret[:32] + b'\0' * (32 - len(decoded_secret))
        decryptor = Cipher(algorithms.AES(decoded_secret), modes.ECB()).decryptor()
        logging.getLogger('cterasdk.direct').debug('Decrypting Encryption Key.')
        decrypted_wrapped_key = utf8_decode(decryptor.update(base64.b64decode(wrapped_key)))
        decrypted_key = ''.join(c for c in decrypted_wrapped_key if c.isprintable())[1:-1]
        return base64.b64decode(decrypted_key)
    except (AssertionError, ValueError, binascii.Error) as error:
        logging.getLogger('cterasdk.direct').error('Could not decrypt secret key. %s', error)
    raise DirectIOError()


def decrypt_block(block, encryption_key):
    try:
        initialization_vector = block[1:17]
        encrypted_data = block[17:]
        logging.getLogger('cterasdk.direct').debug('Decrypting Block.')
        decryptor = Cipher(algorithms.AES(encryption_key), modes.CBC(initialization_vector)).decryptor()
        return decryptor.update(encrypted_data)
    except ValueError as error:
        logging.getLogger('cterasdk.direct').error('Failed to decrypt block. Key error. %s', error)
    except UnsupportedAlgorithm as error:
        logging.getLogger('cterasdk.direct').error('Failed to decrypt block. Unsupported algorithm. %s', error)
    raise DirectIOError()



from ..lib import QueryIterator, DefaultResponse, KeyValueQueryIterator, Command
from ..common import Object


def create_callback_function(edge, path, name=None, *, callback_response=None):
    """
    Create a query callback function

    :param cterasdk.objects.core.Portal core: Portal object
    :param cterasdk.edge.query.QueryParam param: Query paramter object
    :param str,optional name: Schema method name

    :returns: Command object
    """

    def database(edge, path, name, param):
        return edge.api.database(path, name, param)

    def execute(edge, path, name, param):
        response_consumer = callback_response if callback_response else DefaultResponse
        return response_consumer(edge.api.execute(path, name, param))

    return Command(execute if name else database, edge, path, name or 'query')


def iterator(edge, path, param=None, name=None, callback_response=None):
    """
    Create iterator

    :param cterasdk.objects.edge.Edge edge: Edge object
    :param str path: URL Path
    :param str,optional name: Schema method name
    :param cterasdk.edge.query.QueryParam,optional param: Query paramter object
    :param cterasdk.lib.iterator.BaseResponse callback_response: Class to consume callback response

    :returns: Query iterator object
    """
    object_iterator = QueryIterator if name else KeyValueQueryIterator
    callback_function = create_callback_function(edge, path, name, callback_response=callback_response)
    return object_iterator(callback_function, param if param else QueryParam())


class QueryParam(Object):

    def __init__(self):
        self.startFrom = 0
        self.countLimit = 50

    def include_classname(self):
        self._classname = self.__class__.__name__  # pylint: disable=attribute-defined-outside-init

    def increment(self):
        self.startFrom = self.startFrom + self.countLimit


class QueryParamBuilder:

    def __init__(self):
        self.param = QueryParam()

    def startFrom(self, startFrom):
        self.param.startFrom = startFrom
        return self

    def countLimit(self, countLimit):
        self.param.countLimit = countLimit
        return self

    def include(self, include):
        self.param.include = include  # pylint: disable=attribute-defined-outside-init
        return self

    def put(self, key, value):
        setattr(self.param, key, value)
        return self

    def build(self):
        return self.param



import logging
import time

from ..exceptions import CTERAException
from .base_command import BaseCommand


class Power(BaseCommand):
    """ Edge Filer Power APIs """

    def reboot(self, wait=False):
        """
        Reboot the Gateway

        :param bool,optional wait: Wait for reboot to complete, defaults to False
        """
        logging.getLogger('cterasdk.edge').info("Rebooting device. %s", {'host': self._edge.host()})
        self._edge.api.execute("/status/device", "reboot", None)
        if wait:
            Boot(self._edge).wait()

    def shutdown(self):
        """ Shutdown the Edge Filer"""
        self._edge.api.execute("/status/device", "poweroff", None)

    def reset(self, wait=False):
        """
        Reset the Edge Filer setting

        :param bool,optional wait: Wait for reset to complete, defaults to False
        """
        self._edge.api.execute("/status/device", "reset2default", None)
        logging.getLogger('cterasdk.edge').info("Resetting device to default settings. %s", {'host': self._edge.host()})
        if wait:
            Boot(self._edge).wait()


class Boot:

    def __init__(self, edge, retries=60, seconds=5):
        self._edge = edge
        self._retries = retries
        self._seconds = seconds
        self._attempt = 0

    def wait(self):
        while True:
            try:
                self._increment()
                logging.getLogger('cterasdk.edge').debug('Checking if device is up and running. %s', {'attempt': self._attempt})
                self._edge.test()
                logging.getLogger('cterasdk.edge').info("Device is back up and running.")
                break
            except (CTERAException, ConnectionError, TimeoutError) as e:
                logging.getLogger('cterasdk.edge').debug('Exception. %s', {'exception': e.__class__.__name__, 'message': e.message})

    def _increment(self):
        self._attempt = self._attempt + 1
        if self._attempt >= self._retries:
            self._unreachable()
        logging.getLogger('cterasdk.edge').debug('Sleep. %s', {'seconds': self._seconds})
        time.sleep(self._seconds)

    def _unreachable(self):
        host = self._edge.host()
        port = self._edge.port()
        logging.getLogger('cterasdk.edge').error('Timed out. Could not reach host. %s', {'host': host, 'port': port})
        raise ConnectionError(f'Timed out. Could not reach host {host}:{port}.')



import logging
import time

from .base_command import BaseCommand
from .enum import ServerMode, SetupWizardStage, SetupWizardStatus, SlaveAuthenticaionMethod
from ..common import Object, utf8_decode
from ..convert import toxmlstr
from ..clients.common import MultipartForm
from ..exceptions import CTERAException


class Setup(BaseCommand):
    """
    Global Admin Setup APIs
    """

    def __init__(self, portal):
        super().__init__(portal)
        self.stage = None

    def _get_current_stage(self):
        status = self.get_setup_status()
        self.stage = status.wizard
        return self.stage

    def init_master(self, name, email, first_name, last_name, password, domain):
        """
        Initialize the CTERA Portal master server.

        :param str name: User name for the new user
        :param str email: E-mail address of the new user
        :param str first_name: The first name of the new user
        :param str last_name: The last name of the new user
        :param str password: Password for the new user
        :param str domain: The domain suffix for CTERA Portal
        """
        self._get_current_stage()

        params = Setup._init_server_params(ServerMode.Master)
        self._init_server(params, True)

        if self.stage == SetupWizardStage.Portal:
            params = Object()
            params._classname = 'InitParams'  # pylint: disable=protected-access

            params.admin = Object()
            params.admin._classname = 'PortalAdmin'  # pylint: disable=protected-access
            params.admin.name = name
            params.admin.email = email
            params.admin.firstName = first_name
            params.admin.lastName = last_name
            params.admin.password = password

            params.settings = Setup.default_settings()
            params.settings.dnsSuffix = domain
            logging.getLogger('cterasdk.core').info('Initializing Portal. %s', {'domain': domain, 'user': name})
            self._core.ctera.execute('/public', 'init', params)
            SetupWizardStatusMonitor(self._core).wait(SetupWizardStage.Portal)
            logging.getLogger('cterasdk.core').info('Portal initialized.')
        elif self.stage == SetupWizardStage.Finish:
            logging.getLogger('cterasdk.core').warning('Portal already initialized. %s', {'host': self._core.host()})
        self._core.startup.wait()

    def _init_slave(self, ipaddr, secret):
        self._get_current_stage()
        response = self._core.ctera.execute('/setup/authenticaionMethod', 'askMasterForSlaveAuthenticaionMethod', ipaddr)
        params = Setup._init_server_params(ServerMode.Slave)
        params.slaveSettings.masterIpAddr = ipaddr
        if response == SlaveAuthenticaionMethod.Password:
            params.slaveSettings.masterPassword = secret
        elif response == SlaveAuthenticaionMethod.PrivateKey:
            params.slaveSettings.masterKey = secret
        else:
            logging.getLogger('cterasdk.core').error('Unknown authentication method. %s', {'method': response})
        self._init_server(params, True)

    def _init_server(self, params, wait=False):
        form = MultipartForm()
        if self.stage == SetupWizardStage.Server:
            form.add('inputXml', utf8_decode(toxmlstr(params)))
            form.add('serverMode', params.serverMode)

            if params.serverMode == ServerMode.Slave:
                form.add('masterIpAddr', params.slaveSettings.masterIpAddr)

            logging.getLogger('cterasdk.core').info('Initializing server. %s', {'host': self._core.host(), 'mode': params.serverMode})
            self._core.ctera.multipart('/setup', form)
            if wait:
                status = SetupWizardStatusMonitor(self._core).wait(SetupWizardStage.Server)
                self.stage = status.wizard
                logging.getLogger('cterasdk.core').info('Server initialized. %s', {'host': self._core.host(), 'mode': params.serverMode})
        else:
            logging.getLogger('cterasdk.core').warning('Server already initialized. %s', {'host': self._core.host()})

    def init_application_server(self, ipaddr, secret):
        """
        Initialize a CTERA Portal Application Server.

        :param str ipaddr: The CTERA Portal master server IP address
        :param str secret: A password or a PEM-encoded private key
        """
        self._init_slave(ipaddr, secret)
        logging.getLogger('cterasdk.core').info('Initializing Application Server. %s', {'host': ipaddr})
        self._core.startup.wait()

    @staticmethod
    def _init_server_params(mode):
        params = Object()
        params._classname = 'InitServerParams'  # pylint: disable=protected-access
        params.serverMode = mode
        if mode == ServerMode.Slave:
            params.slaveSettings = Object()
            params.slaveSettings._classname = 'SlaveServerSettings'  # pylint: disable=protected-access
        return params

    def get_setup_status(self):
        return self._core.ctera.get('/setup/status')

    @staticmethod
    def default_settings():
        settings = Object()
        settings._classname = 'SystemSettings'  # pylint: disable=protected-access
        settings.smtpSettings = Object()
        settings.smtpSettings._classname = 'SMTPSettings'  # pylint: disable=protected-access
        settings.smtpSettings.smtpHost = 'your.mail.server'
        settings.smtpSettings.smtpPort = 25
        settings.smtpSettings.enableTls = False

        settings.defaultPortalSettings = Object()
        settings.defaultPortalSettings._classname = 'PortalSettings'  # pylint: disable=protected-access
        settings.defaultPortalSettings.mailSettings = Object()
        settings.defaultPortalSettings.mailSettings._classname = 'MailSettings'  # pylint: disable=protected-access
        settings.defaultPortalSettings.mailSettings.sender = 'no-reply@your.domain'
        return settings


class SetupWizardStatusMonitor:

    def __init__(self, portal, retries=60, seconds=5):
        self._core = portal
        self._retries = retries
        self._seconds = seconds
        self._attempt = 0

    def wait(self, stage):
        status = None
        current_stage = stage
        while current_stage == stage:
            try:
                self._increment()
                logging.getLogger('cterasdk.core').debug('Obtaining wizard status. %s', {'attempt': self._attempt})
                status = self._core.setup.get_setup_status()
                logging.getLogger('cterasdk.core').debug('Current wizard status. %s', {
                    'stage': status.wizard,
                    'status': status.currentWizardProgress,
                    'description': status.description
                })
                if status.currentWizardProgress == SetupWizardStatus.Failed:
                    raise CTERAException('Initialization failed.', status)
                current_stage = status.wizard
            except (ConnectionError, TimeoutError) as e:
                logging.getLogger('cterasdk.core').debug('Exception. %s', e.__dict__)
        logging.getLogger('cterasdk.core').debug('Wizard update. %s', {'previous_stage': stage, 'current_stage': current_stage})
        return status

    def _increment(self):
        self._attempt = self._attempt + 1
        if self._attempt >= self._retries:
            SetupWizardStatusMonitor._unreachable()
        logging.getLogger('cterasdk.core').debug('Sleep. %s', {'seconds': self._seconds})
        time.sleep(self._seconds)

    @staticmethod
    def _unreachable():
        logging.getLogger('cterasdk.core').error('Timed out. Setup did not complete in a timely manner.')
        raise CTERAException('Timed out. Setup did not complete in a timely manner')



import copy
from abc import abstractmethod

from .types import ConnectionStatus
from ...common import Object, Version
from ...exceptions import CTERAException


class BaseUser(Object):
    """Base User Account"""

    def __init__(self, name, domain=None):
        self.name = name
        self.domain = domain


class BaseSession(Object):
    """Base CTERA Session"""

    def __init__(self, address, product):
        """
        Initialize a Base Session Object

        :param str address: Hostname or IP address
        :param cterasdk.lib.session.types.Product product: Product
        """
        self.address = address
        self.product = product
        self.connection_status = ConnectionStatus.Disconnected
        self.account = None
        self.software_version = None

    def _before_session_start(self):
        self.connection_status = ConnectionStatus.Connecting

    def start_session(self, session):
        self._before_session_start()
        try:
            self._start_session(session)
            self._activate_connection()
        except CTERAException:
            self._deactivate_connection()
            raise

    @abstractmethod
    def _start_session(self, session):
        return NotImplementedError("Subclass must implement the '_start_session' method.")

    async def async_start_session(self, session):
        self._before_session_start()
        try:
            await self._async_start_session(session)
            self._activate_connection()
        except CTERAException:
            self._deactivate_connection()
            raise

    @abstractmethod
    async def _async_start_session(self, session):
        return NotImplementedError("Subclass must implement the '_start_session_async' method.")

    def _activate_connection(self):
        self._update_connection_status(ConnectionStatus.Connected)

    def stop_session(self):
        self._stop_session()
        self._deactivate_connection()

    def _deactivate_connection(self):
        self._update_connection_status(ConnectionStatus.Disconnected)
        self.account = None
        self.software_version = None

    def _update_account(self, account):
        self.account = account

    def _update_connection_status(self, connection_status):
        self.connection_status = connection_status

    def _update_software_version(self, software_version):
        self.software_version = Version(software_version)

    @property
    def connecting(self):
        return self.connection_status == ConnectionStatus.Connecting

    @property
    def connected(self):
        return self.connection_status == ConnectionStatus.Connected

    def whoami(self):
        session = copy.deepcopy(self)
        session.software_version = str(session.software_version) if session.software_version else None
        print(session)



import os
import errno
import mimetypes
import logging
from pathlib import Path

import aiofiles
import cterasdk.settings


class FileSystem:  # pylint: disable=unused-private-member

    __instance = None

    def __init__(self):
        if FileSystem.__instance is not None:
            raise Exception("FileSystem is a Singleton Class.")
        FileSystem.__instance = self

    @staticmethod
    def instance():
        if FileSystem.__instance is None:
            FileSystem()
        return FileSystem.__instance

    @staticmethod
    def expanduser(p):
        """
        Return a new path with expanded ~ and ~user constructs

        :param str p: Path
        :returns: Absolute Path.
        :rtype: str
        """
        return Path(p).expanduser()

    @staticmethod
    def exists(p):
        """
        Check if a file or a directory exists

        :param str p: Path
        :returns: ``True`` if exists, ``False`` otherwise.
        :rtype: bool
        """
        return Path(p).exists()

    def rename(self, parent, source, destination):  # pylint: disable=no-self-use
        """
        Rename a file or a directory.

        :param str parent: Parent directory
        :param str source: Source file or directory name
        :param str destination: Destination file or directory name
        :returns: Parent directory, and file or directory name
        :rtype: tuple(str, str)
        """
        source = Path(parent).joinpath(source)
        if not source.exists():
            logging.getLogger('cterasdk.filesystem').error('Rename failed. File not found. %s', {'path': source.as_posix()})
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), source.as_posix())
        destination = Path(parent).joinpath(destination)
        source.rename(destination)
        return destination.parent.as_posix(), destination.name

    @staticmethod
    def join(*paths):
        return Path(*paths)

    def is_dir(self, p):
        """
        Check is a directory.

        :param str p: Path
        :returns: ``True`` if a directory, ``False`` otherwise.
        :rtype: bool
        """
        p = self.expanduser(p)
        return p.is_dir()

    def downloads_directory(self):
        """
        Get downloads directory.

        :returns: Directory Path
        :rtype: str
        """
        location = cterasdk.settings.downloads.location
        if not self.is_dir(location):
            logging.getLogger('cterasdk.filesystem').error('Could not find downloads directory. %s', {'path': location})
            raise FileNotFoundError(errno.ENOENT, 'No such directory', location)
        return location

    def split_file_directory(self, location):
        """
        Split file and directory.

        :param str path: Path

        Returns:
        1. (parent directory, file name), if a file exists
        2. (parent directory, file name), if a directory exists
        3. (parent directory, file name), if the parent directory exists
        4. Raises ``FileNotFoundError`` if neither the object nor the parent directory exist
        """
        p = self.expanduser(location)
        if p.exists():
            if p.is_dir():
                filename = None
            else:
                filename = p.name
                p = p.parent
        elif p.parent.exists():
            filename = p.name
            p = p.parent
        else:
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), location)
        return str(p.resolve()), filename

    def generate_file_location(self, location=None, default_filename=None):
        """
        Compute destination file path.

        :param str location: Path to a file or a folder
        :param str default_filename: Default file name, unless ``location`` already specifies a file path
        :returns: Tuple including the destination directory and file name
        :rtype: tuple(str, str)
        """
        parent = filename = None
        if location:
            parent, filename = self.split_file_directory(location)
        else:
            parent = self.downloads_directory()

        if not filename:
            filename = default_filename

        return (parent, filename)

    def properties(self, location):
        """
        Get file properties.

        :param str location: Path
        :returns: File name, size and type
        :rtype: dict
        """
        p = self.expanduser(location)

        if not p.exists():
            logging.getLogger('cterasdk.filesystem').error('File not found. %s', {'path': p.as_posix()})
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), p.as_posix())

        if not p.is_file():
            logging.getLogger('cterasdk.filesystem').error('No such file. %s', {'path': p.as_posix()})
            raise FileNotFoundError(errno.ENOENT, 'Not such file', p.as_posix())

        return dict(name=p.name, size=str(p.stat().st_size), mimetype=mimetypes.guess_type(location))

    @staticmethod
    def file_version(filename, version):
        """
        Append version number to file name.

        :param str filename: File name
        :param int version: File version
        :returns: File name appended with a version number
        :rtype: str
        """
        idx = filename.rfind('.')
        extension = ''
        if idx > 0:
            name = filename[:idx]
            extension = filename[idx:]
        else:
            name = filename
        return f'{name} ({str(version)}){extension}'

    @staticmethod
    def compute_zip_file_name(cloud_directory, files):
        """
        Compute zip file name.
        """
        if len(files) > 1:
            path = Path(cloud_directory)
        else:
            path = Path(files[0])
        return f'{path.stem}.zip'

    def _before_save(self, directory, filename):
        """
        Check directory exists, and compute temporary file name.
        """
        parent = self.expanduser(directory)
        if not parent.exists():
            raise FileNotFoundError(errno.ENOENT, 'No such directory', directory)

        temporary_file = parent.joinpath(f'{filename}.ctera')
        return (parent, temporary_file)

    def _after_write(self, parent, temporary_file, filename):
        """
        Move file to its final destination.
        """
        origin = filename
        version = 0
        while True:
            try:
                self.rename(parent, temporary_file.as_posix(), filename)
                break
            except (FileExistsError, IsADirectoryError):
                logging.getLogger('cterasdk.filesystem').debug('File exists. %s', {'path': parent.as_posix(), 'name': filename})
                version = version + 1
                filename = self.file_version(origin, version)

        filepath = parent.joinpath(filename)
        logging.getLogger('cterasdk.filesystem').info('Saved. %s', {'path': filepath.as_posix()})
        return filepath.as_posix()

    def save(self, directory, filename, handle):
        """
        Save file.

        :param str parent: Directory path.
        :param str filename: File name.
        :param object handle: File handle.
        :returns: File path
        :rtype: str
        """
        parent, temporary_file = self._before_save(directory, filename)  # Check directory exists, and compute temporary file name.
        self.write(temporary_file, handle)  # Write temporary file.
        return self._after_write(parent, temporary_file, filename)  # Rename to destination

    @staticmethod
    def write(p, handle):
        with open(p, 'w+b') as fd:
            if isinstance(handle, bytes):
                fd.write(handle)
            else:
                for chunk in handle.iter_content(chunk_size=8192):
                    fd.write(chunk)
        logging.getLogger('cterasdk.filesystem').debug('Write Complete. %s', {'path': p})

    @staticmethod
    async def async_write(p, handle):
        async with aiofiles.open(p, 'w+b') as fd:
            async for chunk in handle.async_iter_content(chunk_size=8192):
                await fd.write(chunk)
        logging.getLogger('cterasdk.filesystem').debug('Write Complete. %s', {'path': p})



from ..common import Object


class ClientSettings(Object):
    """
    Asynchronous HTTP Client Session Settings
    """

    def __init__(self, connector=None, timeout=None, cookie_jar=None):
        self.connector = connector if connector else TCPConnector(True)
        self.timeout = timeout if timeout else ClientTimeout(None, None)
        self.cookie_jar = cookie_jar if cookie_jar else CookieJar(False)


class TCPConnector(Object):
    """
    Asynchronous HTTP TCP Connector
    """

    def __init__(self, ssl):
        self.ssl = ssl


class ClientTimeout(Object):
    """
    Asynchronous HTTP Client Timeout Settings
    """

    def __init__(self, sock_connect, sock_read):
        self.sock_connect = sock_connect
        self.sock_read = sock_read


class CookieJar(Object):
    """
    Asynchronous HTTP Cookie Jar Settings
    """

    def __init__(self, allow_unsafe):
        self.unsafe = allow_unsafe



import platform


class Platform:  # pylint: disable=unused-private-member

    __instance = None

    @staticmethod
    def instance():
        if Platform.__instance is None:
            Platform()
        return Platform.__instance

    def __init__(self):
        if Platform.__instance is not None:
            raise Exception("Platform is a singleton class.")
        self.hostname = platform.node()
        self.operating_system = platform.system()
        self.release = platform.release()
        self.machine = platform.machine()
        self.python_major, self.python_minor, self.python_micro = platform.python_version_tuple()
        Platform.__instance = self

    def arch(self):
        return self.machine

    def os(self):
        return ' '.join([self.operating_system, str(self.release)])

    def python_version(self):
        return '.'.join([self.python_major, self.python_minor, self.python_micro])



from .item import Item  # noqa: E402, F401
from .object import Object, Device, delete_attrs  # noqa: E402, F401
from .datetime_utils import DateTimeUtils, from_iso_format  # noqa: E402, F401
from .utils import merge, union, parse_base_object_ref, convert_size, df_military_time, DataUnit, parse_to_ipaddress, \
                   utf8_decode, Version  # noqa: E402, F401
from .types import PolicyRule, PolicyRuleConverter, StringCriteriaBuilder, IntegerCriteriaBuilder, DateTimeCriteriaBuilder, \
                   PredefinedListCriteriaBuilder, CustomListCriteriaBuilder, ThrottlingRuleBuilder, ThrottlingRule, FilterBackupSet, \
                   FileFilterBuilder, ApplicationBackupSet, TimeRange  # noqa: E402, F401



from ...common import Object


class Event(Object):  # pylint: disable=too-many-instance-attributes
    """
    :ivar str type: Object type
    :ivar str guid: Object Unique Identifier (GUID)
    :ivar int folder_id: Cloud Drive Folder ID
    :ivar bool deleted: Deleted
    :ivar str name: Object Name
    :ivar str modified: Last Modified Timestamp (ISO 8601)
    :ivar int size: Object Size
    :ivar int,optional id: File ID
    :ivar str,optional acl: ACL SDDL
    :ivar str,optional gvsn: GVSN
    :ivar str,optional parent_guid: Parent Object Unique Identifier (GUID)
    """
    def __init__(self, type, guid, folder_id, deleted, name,  # pylint: disable=redefined-builtin, too-many-arguments
                 modified, size, id=None, acl=None, gvsn=None, parent_guid=None):  # pylint: disable=redefined-builtin
        self.type = type
        self.guid = guid
        self.folder_id = folder_id
        self.deleted = deleted
        self.name = name
        self.modified = modified
        self.size = size
        self.id = id
        self.acl = acl
        self.gvsn = gvsn
        self.parent_guid = parent_guid

    @staticmethod
    def from_server_object(server_object):
        return Event(**server_object.__dict__)



import re
import logging

from . import query
from ..common import Object
from ..lib.task_manager_base import TaskBase
from ..exceptions import InputError
from .base_command import BaseCommand


class Task(TaskBase):

    def _get_task_id(self, ref):
        uid = None
        if isinstance(ref, int):
            uid = str(ref)
        elif isinstance(ref, str):
            match = re.search('[1-9][0-9]*', ref)
            if match is not None:
                start, end = match.span()
                uid = ref[start: end]
        if uid is not None:
            return '/proc/bgtasks/' + uid
        logging.getLogger('cterasdk.edge').error('Could not parse task id. %s', {'ref': ref})
        raise InputError('Invalid task id', ref, [64, '64', '/proc/bgtasks/64'])

    def get_task_status(self):
        return self.CTERAHost.api.get(self.path)


class Tasks(BaseCommand):
    """ Edge Filer Background Task APIs """

    def status(self, ref):
        """
        Get background task status

        :param str ref: Task reference
        """
        task = Task(self._edge, ref)
        return task.status()

    def running(self):
        """
        Get all running background tasks
        """
        return self._query('status', 'running')

    def by_name(self, name):
        """
        Get background tasks by name

        :param str name: Task name
        """
        return self._query('name', name)

    def _query(self, key, value):
        param = Object()
        param.key = key
        param.value = value
        return query.iterator(self._edge, '/proc/bgtasks', param)

    def wait(self, ref, retries=100, seconds=1):
        """
        Wait for background task to complete

        :param str ref: Task reference
        :param int,optional retries: Number of retries when sampling the task status, defaults to 100
        :param int,optional seconds: Number of seconds to wait between retries, defaults to 1
        """
        task = Task(self._edge, ref, retries, seconds)
        return task.wait()



from abc import ABC, abstractmethod

from ..common import utf8_decode
from ..convert import toxmlstr
from .filesystem import FileSystem


class FileAccessBase(ABC):
    def __init__(self, ctera_host):
        self._ctera_host = ctera_host
        self._filesystem = FileSystem.instance()

    def download(self, path, destination=None):
        directory, filename = self._split_destination(destination, path.name)
        handle = self._openfile(path)
        return self._filesystem.save(directory, filename, handle)

    def download_as_zip(self, cloud_directory, files, destination=None):
        files = files if isinstance(files, list) else [files]
        directory, filename = self._split_destination(
            destination,
            self._filesystem.compute_zip_file_name,
            cloud_directory=cloud_directory.fullpath(),
            files=files
        )
        handle = self._get_zip_file_handle(cloud_directory, files)
        return self._filesystem.save(directory, filename, handle)

    def upload(self, local_file, dest_path):
        local_file_info = self._filesystem.properties(local_file)
        with open(local_file, 'rb') as fd:
            return self._upload_object(local_file_info, fd, dest_path)

    @abstractmethod
    def _upload_object(self, local_file_info, fd, dest_path):
        raise NotImplementedError("Subclass must implement _upload_object")

    @abstractmethod
    def _get_upload_url(self, dest_path):
        raise NotImplementedError("Subclass must implement _get_upload_url")

    @abstractmethod
    def _get_upload_form(self, local_file_info, fd, dest_path):
        raise NotImplementedError("Subclass must implement _get_upload_form")

    def _openfile(self, path):
        return self._ctera_host.io.download(self._get_single_file_url(path))

    @abstractmethod
    def _get_single_file_url(self, path):
        raise NotImplementedError("Subclass must implement _get_single_file_url")

    def _get_zip_file_handle(self, cloud_directory, files):
        raise NotImplementedError("Subclass must implement _get_zip_file_handle")

    @abstractmethod
    def _get_multi_file_url(self, cloud_directory, files):
        raise NotImplementedError("Subclass must implement _get_multi_file_url")

    @property
    def _use_file_url_for_multi_file_url(self):
        raise NotImplementedError("Subclass must implement _use_file_url_for_multi_file_url")

    def _make_form_data(self, cloud_directory, files):
        return dict(
            inputXML=utf8_decode(toxmlstr(self._get_multi_file_object(cloud_directory, files)))
        )

    @abstractmethod
    def _get_multi_file_object(self, cloud_directory, files):
        raise NotImplementedError("Subclass must implement _get_multi_file_object")

    def _split_destination(self, destination, filename_resolver, **filename_resolver_kwargs):
        directory = filename = None
        if destination:
            directory, filename = self._filesystem.split_file_directory(destination)
        else:
            directory = self._filesystem.downloads_directory()
        if filename is None:
            filename = filename_resolver(**filename_resolver_kwargs)
        return directory, filename



import logging

from .base_command import BaseCommand
from ..lib import X509Certificate, PrivateKey, create_certificate_chain
from ..common import Object


def initialize(edge):
    """
    Conditional intialization of the Edge Filer SSL Module.
    """
    if edge.session().software_version > '7.8':
        return SSLv78(edge)
    return SSLv1(edge)


class SSL(BaseCommand):
    """ Edge Filer SSL APIs """

    def enable_http(self):
        """
        Enable HTTP access
        """
        self._set_force_https(False)

    def disable_http(self):
        """
        Disable HTTP access
        """
        self._set_force_https(True)

    def is_http_disabled(self):
        """
        Check if HTTP access is disabled
        """
        return self._get_force_https()

    def is_http_enabled(self):
        """
        Check if HTTP access is enabled
        """
        return not self._get_force_https()

    def _get_force_https(self):
        return self._edge.api.get('/config/fileservices/webdav/forceHttps')

    def _set_force_https(self, force):
        self._edge.api.put('/config/fileservices/webdav/forceHttps', force)

    def _import_certificate(self, path, private_key, *certificates):
        key_object = PrivateKey.load_private_key(private_key)
        certificates = [X509Certificate.load_certificate(certificate) for certificate in certificates]
        certificate_chain = [certificate.pem_data.decode('utf-8') for certificate in create_certificate_chain(*certificates)]
        server_certificate = ''.join([key_object.pem_data.decode('utf-8')] + certificate_chain)
        logging.getLogger('cterasdk.edge').info("Uploading Web Server SSL certificate.")
        response = self._edge.api.put(path, f"\n{server_certificate}")
        logging.getLogger('cterasdk.edge').info("Uploaded Web Server SSL certificate.")
        return response


class SSLv1(SSL):
    """ Edge Filer SSLv1 APIs """

    def remove_storage_ca(self):
        """
        Remove object storage trusted CA certificate
        """
        self._edge.api.put('/config/extStorageTrustedCA', None)

    def get_storage_ca(self):
        """
        Get object storage trusted CA certificate
        """
        return self._edge.api.get('/status/extStorageTrustedCA')

    def import_storage_ca(self, certificate):
        """
        Import the object storage trusted CA certificate

        :param str certificate: The PEM-encoded certificate or a path to the PEM-encoded server certificate file
        """
        logging.getLogger('cterasdk.edge').info('Setting trusted object storage CA certificate')
        param = Object()
        param._classname = 'ExtTrustedCA'  # pylint: disable=protected-access
        param.certificate = X509Certificate.load_certificate(certificate).pem_data.decode('utf-8')
        logging.getLogger('cterasdk.edge').info("Uploading object storage certificate.")
        response = self._edge.api.put('/config/extStorageTrustedCA', param)
        logging.getLogger('cterasdk.edge').info("Uploaded object storage certificate.")
        return response

    def import_certificate(self, private_key, *certificates):
        """
        Import the Edge Filer's web server's SSL certificate

        :param str private_key: The PEM-encoded private key, or a path to the PEM-encoded private key file
        :param list[str] certificates: The PEM-encoded certificates, or a list of paths to the PEM-encoded certificates
        """
        return self._import_certificate('/config/certificate', private_key, *certificates)


class SSLv78(SSL):
    """ Edge Filer v7.8 SSL Certificate APIs """

    def __init__(self, edge):
        super().__init__(edge)
        self.server = ServerCertificate(edge)
        self.ca = TrustedCAs(edge)


class ServerCertificate(BaseCommand):
    """ Edge Filer v7.8 Server Certificate APIs """

    def get(self):
        """
        Get Server Cerificate.
        """
        return self._edge.api.get('/proc/certificates/serverCertificate')

    def regenerate(self):
        """
        Generate a Self Signed Certificate.
        """
        logging.getLogger('cterasdk.edge').info("Generating a Self Signed Certificate.")
        response = self._edge.api.execute('/config/certificates', 'createSelfSign')
        logging.getLogger('cterasdk.edge').info("Generated a Self Signed Certificate.")
        return response

    def import_certificate(self, private_key, *certificates):
        """
        Import the Edge Filer's server SSL certificate

        :param str private_key: The PEM-encoded private key, or a path to the PEM-encoded private key file
        :param list[str] certificates: The PEM-encoded certificates, or a list of paths to the PEM-encoded certificates
        """
        return self._edge.ssl._import_certificate('/config/certificates/serverCertificate',  # pylint: disable=protected-access
                                                  private_key, *certificates)


class TrustedCAs(BaseCommand):
    """ Edge Filer v7.8 Trusted CAs APIs """

    def all(self):
        """
        List Trusted CAs.
        """
        return self._edge.api.get('/proc/certificates/trustedCACertificates')

    def add(self, ca):
        """
        Add Trusted CA.

        :param str certificate: The PEM-encoded certificate or a path to the PEM-encoded server certificate file
        """
        certificate = X509Certificate.load_certificate(ca).pem_data.decode('utf-8')
        return self._edge.api.execute('/config/certificates', 'addTrustedCACert', certificate)

    def remove(self, ca):
        """
        Remove Trusted CA.

        :param object ca: CA fingerprint as `str`, or a trusted CA object.
        """
        fingerprint = None
        if isinstance(ca, Object) and hasattr(ca, 'fingerprint'):
            fingerprint = ca.fingerprint
        else:
            fingerprint = ca

        if fingerprint:
            logging.getLogger('cterasdk.edge').info("Removing Trusted CA. %s", {'fingerprint': fingerprint})
            response = self._edge.api.delete(f'/config/certificates/trustedCACertificates/{fingerprint}')
            logging.getLogger('cterasdk.edge').info("Removed Trusted CA. %s", {'fingerprint': fingerprint})
            return response

        raise ValueError('Could not identify CA fingerprint.')

    def clear(self):
        """
        Remove all Trusted CAs.
        """
        logging.getLogger('cterasdk.edge').info("Removing all Trusted CAs.")
        for ca in self.all():
            self.remove(ca)



import logging

from .base_command import BaseCommand


class CLI(BaseCommand):
    """ Portal CLI APIs """

    def run_command(self, cli_command):
        """
        Run a CLI command

        :param str cli_command: The CLI command to run on the gateway
        :return str: The response of the Portal
        """
        logging.getLogger('cterasdk.core').warning('Usage of the CLI module is discouraged. '
                                                   'Review available modules to determine if there are existing ones that '
                                                   'support this action.')
        logging.getLogger('cterasdk.core').info("Executing CLI command. %s", {'cli_command': cli_command})
        response = self._core.api.execute('', 'debugCmd', cli_command)
        logging.getLogger('cterasdk.core').info("CLI command executed. %s", {'cli_command': cli_command})
        return response



# pylint: disable=wrong-import-position
import cterasdk.settings  # noqa: E402, F401
import cterasdk.logging  # noqa: E402, F401

from .common import Object, PolicyRule  # noqa: E402, F401
from .convert import fromjsonstr, tojsonstr, fromxmlstr, toxmlstr  # noqa: E402, F401
from .exceptions import CTERAException  # noqa: E402, F401
from .core import query  # noqa: E402, F401
from .edge import types as edge_types  # noqa: E402, F401
from .edge import enum as edge_enum  # noqa: E402, F401
from .core import types as core_types  # noqa: E402, F401
from .core import enum as core_enum  # noqa: E402, F401
from .common import types as common_types  # noqa: E402, F401
from .common import enum as common_enum  # noqa: E402, F401
from .objects import GlobalAdmin, ServicesPortal, Edge, Drive, AsyncGlobalAdmin, AsyncServicesPortal  # noqa: E402, F401
from . import direct as ctera_direct  # noqa: E402, F401



from . import uri


class DefaultBuilder:

    def __init__(self, base=None):
        self._base = base

    def __call__(self, url):
        return uri.join(self._base, url)


class EndpointBuilder(DefaultBuilder):

    def __init__(self, base, *segments):
        super().__init__(base)
        self._segments = segments

    def __call__(self, *segments):
        return uri.create(self._base, *(self._segments + segments))

    @staticmethod
    def new(base, *segments):
        return EndpointBuilder(base, *segments)



import logging

from .base_command import BaseCommand
from ..common import Object
from ..exceptions import CTERAException
from .enum import PortalAccountType, SearchType, DirectoryServiceType, DirectoryServiceFetchMode, Role, DirectorySearchEntityType
from .types import AccessControlEntry, AccessControlRule, UserAccount, GroupAccount


class DirectoryService(BaseCommand):
    """
    Portal Active Directory APIs
    """

    def _get_configuration(self):
        return self._core.api.get('/directoryConnector')

    def connected(self):
        directory_services_config = self._get_configuration()
        if directory_services_config is None:
            return False
        try:
            self._connect_to_directory_services(directory_services_config)
            return True
        except CTERAException:
            return False

    # pylint: disable=too-many-arguments
    def connect(self, domain, username, password, directory=DirectoryServiceType.Microsoft,
                domain_controllers=None, ou=None, ssl=False, krb=False, mapping=None, acl=None,
                default=Role.Disabled, fetch=DirectoryServiceFetchMode.Lazy):
        """
        Connect a Portal tenant to directory services

        :param str domain: The directory services domain to connect to
        :param str username: The user name to use when connecting to the active directory services
        :param str password: The password to use when connecting to the active directory services
        :param str,optional ou: The OU path to use when connecting to the active directory services, defaults to `None`
        :param cterasdk.core.enum.DirectoryServiceType,optional directory: The directory service type, deafults to `'ActiveDirectory'`
        :param cterasdk.core.types.DomainControllers,optional domain_controllers:
            Connect to a primary and a secondary domain controllers, defaults to `None`
        :param bool,optional ssl: Connect using SSL, defaults to `False`
        :param bool,optional krb: Connect using Kerberos, defaults to `False`
        :param list[cterasdk.common.types.ADDomainIDMapping],optional: The directory services UID/GID mapping
        :param list[cterasdk.core.types.AccessControlEntry],optional acl: List of access control entries and their associated roles
        :param cterasdk.core.enum.Role default: Default role if no match applies, defaults to `None`
        :param str,optional fetch: Configure identity fetching mode, defaults to `'Lazy'`
        """
        param = Object()
        param._classname = 'ActiveDirectory'  # pylint: disable=protected-access
        param.type = directory
        param.domain = domain
        param.useKerberos = krb
        param.useSSL = ssl
        param.username = username
        param.password = password
        param.ou = ou
        param.noMatchRole = default
        param.accessControlRules = None
        param.idMapping = None
        param.fetchMode = fetch
        param.ipAddresses = None

        if domain_controllers:
            param.ipAddresses = Object()
            param.ipAddresses._classname = 'DomainControlIPAddresses'  # pylint: disable=protected-access
            param.ipAddresses.ipAddress1 = domain_controllers.primary
            param.ipAddresses.ipAddress2 = domain_controllers.secondary

        tenant = self._core.session().account.tenant
        logging.getLogger('cterasdk.core').info('Connecting Portal to directory services. %s', {
            'tenant': tenant,
            'type': type,
            'domain': domain
        })
        self._connect_to_directory_services(param)
        logging.getLogger('cterasdk.core').info('Connected Portal to directory services. %s', {'tennat': tenant, 'domain': domain})

        if mapping:
            self._configure_advanced_mapping(mapping)

        if acl:
            self._configure_access_control(acl, default)

    def _connect_to_directory_services(self, param):
        return self._core.api.execute('', 'testAndSaveAD', param)

    def get_advanced_mapping(self):
        """
        Retrieve directory services advanced mapping configuration

        :returns: A dictionary of domain mapping objects
        :rtype: dict
        """
        return {map.domainFlatName: map for map in self._core.api.get('/directoryConnector/idMapping/map')}

    def set_advanced_mapping(self, mapping):
        """
        Configure advanced mapping

        :param list[cterasdk.common.types.ADDomainIDMapping] mapping: List of domains and their UID/GID mapping range
        """
        if self._get_configuration() is None:
            raise CTERAException('Failed to apply mapping. Not connected to directory services.')

        return self._configure_advanced_mapping(mapping)

    def _configure_advanced_mapping(self, mapping):
        param = Object()
        param._classname = 'ADIDMapping'  # pylint: disable=protected-access
        param.map = mapping
        logging.getLogger('cterasdk.core').debug('Updating advanced mapping. %s', {
            'domains': [mapping.domainFlatName for mapping in param.map]
        })
        response = self._core.api.put('/directoryConnector/idMapping', param)
        logging.getLogger('cterasdk.core').info('Updated advanced mapping.')
        return response

    def get_access_control(self):
        """
        Retrieve directory services access control list

        :returns: List of access control entries
        :rtype: list[cterasdk.core.types.AccessControlEntry]
        """
        acl = []
        for ace in self._core.api.get('/directoryConnector/accessControlRules'):
            if ace.group.type == DirectorySearchEntityType.User:
                acl.append(AccessControlEntry(UserAccount(ace.group.name, ace.group.domain), ace.role))
            elif ace.group.type == DirectorySearchEntityType.Group:
                acl.append(AccessControlEntry(GroupAccount(ace.group.name, ace.group.domain), ace.role))
        return acl

    def set_access_control(self, acl=None, default=None):
        """
        Configure directory services access control

        :param list[cterasdk.core.types.AccessControlEntry],optional acl:
            List of access control entries and their associated roles
        :param cterasdk.core.enum.Role default: Default role if no match applies, defaults to `None`
        """
        directory_services_config = self._get_configuration()
        if directory_services_config is None:
            raise CTERAException('Failed to apply access control. Not connected to directory services.')

        default = default if default is not None else directory_services_config.noMatchRole
        return self._configure_access_control(acl, default)

    def _configure_access_control(self, acl, default=None):

        access_control_rules = []
        for ace in acl:
            account = None
            if ace.account.account_type == PortalAccountType.User:
                account = self._search_users(ace.account.directory, ace.account.name)
            elif ace.account.account_type == PortalAccountType.Group:
                account = self._search_groups(ace.account.directory, ace.account.name)
            access_control_rules.append(AccessControlRule(account, ace.role))

        logging.getLogger('cterasdk.core').info('Updating access control rules.')
        response = self._core.api.put('/directoryConnector/accessControlRules', access_control_rules)
        logging.getLogger('cterasdk.core').info('Updated access control rules.')

        if default is not None:
            logging.getLogger('cterasdk.core').info('Updating default role.')
            response = self._core.api.put('/directoryConnector/noMatchRole', default)
            logging.getLogger('cterasdk.core').info('Updated default role')

        return response

    def get_default_role(self):
        """
        Retrieve the default role assigned when no access control entry match was found
        """
        return self._core.api.get('/directoryConnector/noMatchRole')

    def get_connected_domain(self):
        """
        Get the connected domain information. Returns `None` if the Portal tenant is not connected to a domain

        :return str: The connected domain
        """
        domain = None
        try:
            domain = self._core.api.get('/directoryConnector/domain')
        except CTERAException:
            pass
        return domain

    def domains(self):
        """
        Get domains

        :return list(str): List of names of all discovered domains
        """
        return self._core.api.execute('', 'getADTrustedDomains', False)

    def fetch(self, active_directory_accounts):
        """
        Instruct the Portal to fetch the provided Active Directory Accounts

        :param list[cterasdk.core.types.PortalAccount] active_directory_accounts: List of Active Directory Accounts to fetch

        :return: Response Code
        """
        domains = self._core.domains.list_domains()
        account_types = [v for k, v in PortalAccountType.__dict__.items() if not k.startswith('_')]

        param = []
        for active_directory_account in active_directory_accounts:
            if active_directory_account.directory not in domains:
                logging.getLogger('cterasdk.core').error('Invalid domain name. %s', {'domain': active_directory_account.directory})
                raise CTERAException('Invalid domain', None, domain=active_directory_account.directory, domains=domains)

            if active_directory_account.account_type not in account_types:
                logging.getLogger('cterasdk.core').error('Invalid account type. %s', {'type': active_directory_account.account_type})
                raise CTERAException('Invalid account type', None, type=active_directory_account.account_type, options=account_types)

        for active_directory_account in active_directory_accounts:
            if active_directory_account.account_type == PortalAccountType.User:
                param.append(self._search_users(active_directory_account.directory, active_directory_account.name))
            elif active_directory_account.account_type == PortalAccountType.Group:
                param.append(self._search_groups(active_directory_account.directory, active_directory_account.name))

        logging.getLogger('cterasdk.core').info('Starting to fetch users and groups.')
        response = self._core.api.execute('', 'syncAD', param)
        logging.getLogger('cterasdk.core').info('Started fetching users and groups.')

        return response

    def _search_users(self, domain, user):
        return self._search_directory_services(SearchType.Users, domain, user)

    def _search_groups(self, domain, group):
        return self._search_directory_services(SearchType.Groups, domain, group)

    def _search_directory_services(self, search_type, domain, name):
        param = Object()
        param.mode = search_type
        param.name = name
        param.domain = domain

        logging.getLogger('cterasdk.core').info(
            'Searching %(search_type)s: %(info)s',
            dict(search_type=search_type, info={'domain': domain, 'name': name})
        )

        objects = self._core.api.execute('', 'searchAD', param)
        if not objects:
            logging.getLogger('cterasdk.core').info('Could not find results that match your search criteria. %s',
                                                    {'domain': domain, 'name': name})
            raise CTERAException(
                'Could not find results that match your search criteria',
                None,
                domain=domain,
                type=search_type,
                account=name
            )

        for principal in objects:
            if principal.name == name:
                logging.getLogger('cterasdk.core').info('Found. %s', {'domain': domain, 'name': name})
                return principal

        raise CTERAException(
            'Search returned multiple results, but none matched your search criteria',
            None,
            domain=domain,
            type=search_type,
            account=name
        )

    def disconnect(self):
        """
        Disconnect a Portal tenant from directory services
        """
        return self._core.api.put('/directoryConnector', None)



import logging
from .base import BaseSession, BaseUser
from .types import Product


class PortalUser(BaseUser):
    """Local User"""

    def __init__(self, name, domain, tenant, role):
        super().__init__(name, domain)
        self.tenant = tenant
        self.role = role


class Session(BaseSession):

    Administration = 'Administration'

    def __init__(self, address, context):
        super().__init__(address, Product.Portal)
        self.context = context

    def _start_session(self, session):
        logging.getLogger('cterasdk.core').debug('Starting Session.')
        current_session = session.api.get('/currentSession')
        current_tenant = session.api.get('/currentPortal') or Session.Administration
        software_version = session.api.get('/version')
        self._update_session(current_session, current_tenant, software_version)

    async def _async_start_session(self, session):
        logging.getLogger('cterasdk.core').debug('Starting Session.')
        current_session = session.v1.api.get('/currentSession')
        current_tenant = session.v1.api.get('/currentPortal')
        software_version = session.v1.api.get('/version')
        self._update_session(await current_session, await current_tenant or Session.Administration, await software_version)

    def _update_session(self, current_session, current_tenant, software_version):
        self._update_account(
            PortalUser(
                current_session.username,
                current_session.domain,
                current_tenant,
                current_session.role
            )
        )
        self._update_software_version(software_version)

    def _stop_session(self):  # pylint: disable=no-self-use
        logging.getLogger('cterasdk.core').debug('Stopping Session.')

    def current_tenant(self):
        return self.account.tenant if self.account else None

    def update_current_tenant(self, current_tenant):
        self.account.tenant = current_tenant or Session.Administration

    def in_tenant_context(self):
        return self.current_tenant() != Session.Administration



import re
import logging

from pathlib import PurePosixPath

from .. import query
from ...lib import FetchResourcesResponse
from ...common import Object
from ...objects import uri
from ...exceptions import RemoteDirectoryNotFound, CTERAException


class ItemExists(CTERAException):
    """Already exists"""


class InvalidPath(CTERAException):
    """Invalid directory path"""


class InvalidName(CTERAException):
    """Invalid directory name"""


class ReservedName(CTERAException):
    """Reserved name"""


class SrcDstParam(Object):

    __instance = None

    @staticmethod
    def instance(src, dest=None):
        SrcDstParam(src, dest)
        return SrcDstParam.__instance

    def __init__(self, src, dest=None):
        self._classname = self.__class__.__name__
        self.src = src
        self.dest = dest
        SrcDstParam.__instance = self  # pylint: disable=unused-private-member


class ActionResourcesParam(Object):

    __instance = None

    @staticmethod
    def instance():
        ActionResourcesParam()
        return ActionResourcesParam.__instance

    def __init__(self):
        self._classname = self.__class__.__name__
        self.urls = []
        ActionResourcesParam.__instance = self  # pylint: disable=unused-private-member

    def add(self, param):
        self.urls.append(param)


class CreateShareParam(Object):

    __instance = None

    @staticmethod
    def instance(path, access, expire_on):
        CreateShareParam(path, access, expire_on)
        return CreateShareParam.__instance

    def __init__(self, path, access, expire_on):
        self._classname = self.__class__.__name__
        self.url = path
        self.share = Object()
        self.share._classname = 'ShareConfig'
        self.share.accessMode = access
        self.share.protectionLevel = 'publicLink'
        self.share.expiration = expire_on
        self.share.invitee = Object()
        self.share.invitee._classname = 'Collaborator'
        self.share.invitee.type = 'external'
        CreateShareParam.__instance = self  # pylint: disable=unused-private-member


class FetchResourcesParam(Object):

    def __init__(self):
        self._classname = 'FetchResourcesParam'
        self.start = 0
        self.limit = 100

    def increment(self):
        self.start = self.start + self.limit


class FetchResourcesParamBuilder:

    def __init__(self):
        self.param = FetchResourcesParam()

    def root(self, root):
        self.param.root = root  # pylint: disable=attribute-defined-outside-init
        return self

    def depth(self, depth):
        self.param.depth = depth  # pylint: disable=attribute-defined-outside-init
        return self

    def include_deleted(self):
        self.param.includeDeleted = True  # pylint: disable=attribute-defined-outside-init

    def build(self):
        return self.param


class Path:

    def __init__(self, param, base):
        self._base = PurePosixPath(base)
        self._relative = PurePosixPath()
        if Path._is_server_object(param):
            self._from_server_object(param)
        elif isinstance(param, str):
            self._from_string(param)
        else:
            message = 'Invalid directory path specified. Please ensure the directory path is correct and try again.'
            logging.getLogger('cterasdk.core').error(message)
            raise ValueError(message)

        if self._relative.root == '/' or self._base.joinpath(self._relative) == self._relative:
            raise ValueError('You must specify a relative path. Omit leading / characters')

    @staticmethod
    def _is_server_object(param):
        return isinstance(param, Object) and param.__dict__.get('_classname', None) == 'ResourceInfo'

    def _from_server_object(self, param):
        href = uri.unquote(param.href)
        match = re.search('^/(ServicesPortal|admin)/webdav', href)
        start, end = match.span()
        self._base = self._base.joinpath(href[start: end])
        self._relative = self._relative.joinpath(href[end + 1:])

    def _from_string(self, param):
        self._relative = self._relative.joinpath(param)

    def name(self):
        return self._relative.name

    @property
    def base(self):
        return str(self._base)

    @property
    def relative(self):
        return self._relative

    def parent(self):
        return Path(str(self._relative.parent), str(self._base))

    def fullpath(self):
        return str(self._base.joinpath(self._relative))

    def encoded_fullpath(self):
        return uri.quote(self.fullpath())

    def encoded_parent(self):
        return uri.quote(str(self.parent()))

    def joinpath(self, path):
        return Path(str(self._relative.joinpath(path)), str(self._base))

    def parts(self):
        return self._relative.parts

    def __str__(self):
        return self.fullpath()


def fetch_resources(core, param):
    return core.api.execute('', 'fetchResources', param)


def objects_iterator(core, param):
    return query.iterator(core, '', param, 'fetchResources', callback_response=FetchResourcesResponse)


def get_resource_info(core, path):
    response = core.files.listdir(str(path.relative), depth=0)
    if response.root is None:
        raise RemoteDirectoryNotFound(path.fullpath())
    return response.root


def get_object_path(base, elements):
    if isinstance(elements, list):
        return [Path(element, base) for element in elements]
    return Path(elements, base)


def get_create_dir_param(name, parent):
    param = Object()
    param.name = name
    param.parentPath = parent
    return param


def raise_for_status(response, path):
    error = {
        "FileWithTheSameNameExist": ItemExists(),
        "DestinationNotExists": InvalidPath(),
        "InvalidName": InvalidName(),
        "ReservedName": ReservedName()
    }.get(response, None)
    try:
        if error:
            raise error
    except ItemExists as error:
        logging.getLogger('cterasdk.core').info('A file or folder with the same name already exists. %s', {'path': path})
        raise error
    except InvalidPath as error:
        logging.getLogger('cterasdk.core').error('Invalid parent directory path. %s', {'path': path})
        raise error
    except InvalidName as error:
        logging.getLogger('cterasdk.core').error('Directory name contains invalid characters. %s', {'name': path})
        raise error
    except ReservedName as error:
        logging.getLogger('cterasdk.core').error('Reserved directory name. %s', {'name': path})
        raise error



import logging
from ..lib import FileSystem, CryptoServices
from ..common import Object
from .base_command import BaseCommand


class SSH(BaseCommand):
    """ Edge Filer SSH daemon APIs """

    def enable(self, public_key=None, public_key_file=None, exponent=65537, key_size=2048):
        """
        Enable the Edge Filer's SSH daemon

        :param str,optional public_key: A PEM-encoded public key in OpenSSH format.
         If neither a public key nor public key file were specified, an RSA key pair will be generated automatically.
         The PEM-encoded private key will be saved to the default Downloads directory
        :param str,optional public_key_file: A path to the public key file
        :param int,optional exponent: The public exponent of the new key, defaults to `65537`
        :param int,optional key_size: The length of the modulus in bits, defaults to `2048`
        """
        param = Object()

        if public_key is None:
            if public_key_file is not None:
                FileSystem.instance().properties(public_key_file)
                with open(public_key_file, 'r', encoding='utf-8') as f:
                    public_key = f.read()
            else:
                public_key = CryptoServices.generate_and_save_key_pair(self._edge.host(), exponent=exponent, key_size=key_size)

        param.publicKey = public_key

        logging.getLogger('cterasdk.edge').info("Enabling SSH daemon.")
        self._edge.api.execute('/config/device', 'startSSHD', param)
        logging.getLogger('cterasdk.edge').info("SSH daemon enabled.")

    def disable(self):
        logging.getLogger('cterasdk.edge').info("Disabling SSH daemon.")
        self._edge.api.execute('/config/device', 'stopSSHD')
        logging.getLogger('cterasdk.edge').info("SSH daemon disabled.")



import logging

from ..lib.task_manager_base import TaskError
from ..exceptions import CTERAException
from .base_command import BaseCommand


class Shell(BaseCommand):
    """ Edge Filer Shell command """

    def run_command(self, shell_command, wait=True):
        """
        Execute a Shell Command.

        :param str shell_command: The shell command to execute
        :param bool,optional wait: Wait for the command to execute, defaults to ``True``
        :return: The command result, or the task url path when wait equals ``False``
        """
        logging.getLogger('cterasdk.edge').info("Executing shell command. %s", {'shell_command': shell_command})

        task = self._edge.api.execute("/config/device", "bgshell", shell_command)
        if not wait:
            return task
        try:
            task = self._edge.tasks.wait(task)
            logging.getLogger('cterasdk.edge').info("Shell command executed. %s", {'shell_command': shell_command})
            return task.result.result
        except TaskError as error:
            raise CTERAException('An error occurred while executing task', error)



import logging
from ..common.utils import tcp_connect


def test(Portal):
    tcp_connect(Portal.host(), Portal.port())
    logging.getLogger('cterasdk.core').debug('Trying to obtain Portal public info.')
    response = Portal.public_info()
    logging.getLogger('cterasdk.core').debug('Successfully obtained Portal public info. %s',
                                             {'version': response.version, 'name': response.name})
    return response



import logging
from .base_command import BaseCommand
from . import enum
from ..exceptions import InputError


class Reports(BaseCommand):
    """
    Reports APIs
    """
    def storage(self):
        """
        Retrieve the portals statistics report.\n
        To retrieve this report, you must first browse the Global Administration Portal, using: `GlobalAdmin.portals.browse_global_admin()`
        """
        return self._get_report(enum.Reports.Storage)

    def portals(self):
        """
        Retrieve the storage statistics report.\n
        To retrieve this report, you must first browse the Global Administration Portal, using: `GlobalAdmin.portals.browse_global_admin()`
        """
        return self._get_report(enum.Reports.Portals)

    def folders(self):
        """
        Retrieve the cloud folders statistics report.\n
        To retrieve this report, you must first browse the Virtual Portal that contains the report, using: `GlobalAdmin.portals.browse()`
        """
        return self._get_report(enum.Reports.Folders)

    def folder_groups(self):
        """
        Retrieve the folder groups statistics report.\n
        To retrieve this report, you must first browse the Virtual Portal that contains the report, using: `GlobalAdmin.portals.browse()`
        """
        return self._get_report(enum.Reports.FolderGroups)

    def generate(self, name):
        """
        Generate a CTERA Portal Report

        :param cterasdk.core.enum.reports report: Report
        """
        options = {v: k for k, v in enum.Reports.__dict__.items() if not k.startswith('_')}
        if options.get(name, None) is None:
            raise InputError('Invalid report type', name, list(options.values()))
        logging.getLogger('cterasdk.core').info('Running Portal Report. %s', {'name': name})
        self._core.api.execute('', 'generateReport', name)

    def _get_report(self, report_name):
        return self._core.api.get(f'/reports/{report_name}')




from . import common
from ...common import Object
from ...lib import FileAccessBase
from ...exceptions import RemoteFileSystemException


class FileAccess(FileAccessBase):

    def _get_single_file_url(self, path):
        return path.relative

    def _get_multi_file_url(self, cloud_directory, files):
        return str(self._get_cloud_folder_uid(cloud_directory))

    @property
    def _use_file_url_for_multi_file_url(self):
        return True

    def _get_multi_file_object(self, cloud_directory, files):
        files_obj = Object()
        files_obj.paths = ['/'.join([cloud_directory.fullpath(), filename]) for filename in files]
        files_obj.snapshot = None
        files_obj.password = None
        files_obj.portalName = None
        files_obj.showDeleted = False
        return files_obj

    def _get_upload_url(self, dest_path):
        return str(self._get_cloud_folder_uid(dest_path))

    def _get_upload_form(self, local_file_info, fd, dest_path):
        return dict(
            name=local_file_info['name'],
            Filename=local_file_info['name'],
            fullpath=self._ctera_host.io.builder(common.get_object_path(dest_path.relative,  # pylint: disable=protected-access
                                                                        local_file_info['name']).encoded_fullpath()),
            fileSize=local_file_info['size'],
            file=fd
        )

    def _get_cloud_folder_uid(self, path):
        resource_info = common.get_resource_info(self._ctera_host, path)
        if not resource_info.isFolder:
            raise RemoteFileSystemException('The destination path is not a directory', None, path=path.fullpath())
        return resource_info.cloudFolderInfo.uid

    def _get_zip_file_handle(self, cloud_directory, files):
        return self._ctera_host.io.download_zip(  # pylint: disable=protected-access
            self._get_multi_file_url(cloud_directory, files),
            self._make_form_data(cloud_directory, files)
        )

    def _upload_object(self, local_file_info, fd, dest_path):
        return self._ctera_host.io.upload(  # pylint: disable=protected-access
                self._get_upload_url(dest_path),
                self._get_upload_form(local_file_info, fd, dest_path)
            )



from abc import abstractmethod
from ...clients.synchronous import clients
from ..services import Management
from ..endpoints import EndpointBuilder

from .. import authenticators

from ...lib.session.core import Session

from ...core import activation
from ...core import admins
from ...core import antivirus
from ...core import buckets
from ...core import cli
from ...core import cloudfs
from ...core import connection
from ...core import credentials
from ...core import devices
from ...core import directoryservice
from ...core import domains
from ...core import files
from ...core import firmwares
from ...core import groups
from ...core import kms
from ...core import licenses
from ...core import login
from ...core import logs
from ...core import mail
from ...core import messaging
from ...core import plans
from ...core import portals
from ...core import reports
from ...core import roles
from ...core import servers
from ...core import settings
from ...core import setup
from ...core import ssl
from ...core import startup
from ...core import storage_classes
from ...core import syslog
from ...core import taskmgr
from ...core import templates
from ...core import users


class Clients:

    def __init__(self, core):
        self.ctera = core.default.clone(clients.Extended, EndpointBuilder.new(core.base, core.context))
        self.api = core.default.clone(clients.API, EndpointBuilder.new(core.base, core.context, '/api'))
        self.io = IO(core)


class IO:

    def __init__(self, core):
        self._folders = core.default.clone(clients.Folders, EndpointBuilder.new(core.base, core.context, '/folders/folders'))
        self._upload = core.default.clone(clients.Upload, EndpointBuilder.new(core.base, core.context, '/upload/folders'))
        self._webdav = core.default.clone(clients.WebDAV, EndpointBuilder.new(core.base, core.context, '/webdav'))

    @property
    def upload(self):
        return self._upload.upload

    @property
    def download(self):
        return self._webdav.download

    @property
    def download_zip(self):
        return self._folders.download_zip

    @property
    def builder(self):
        return self._webdav._builder  # pylint: disable=protected-access


class Portal(Management):  # pylint: disable=too-many-instance-attributes

    def __init__(self, host, port=None, https=True):
        super().__init__(host, port, https, base=None)

        self._ctera_session = Session(self.host(), self.context)
        self._ctera_clients = Clients(self)

        self.activation = activation.Activation(self)
        self.admins = admins.Administrators(self)
        self.backups = files.Backups(self)
        self.cloudfs = cloudfs.CloudFS(self)
        self.credentials = credentials.Credentials(self)
        self.devices = devices.Devices(self)
        self.directoryservice = directoryservice.DirectoryService(self)
        self.domains = domains.Domains(self)
        self.files = files.CloudDrive(self)
        self.firmwares = firmwares.Firmwares(self)
        self.groups = groups.Groups(self)
        self.logs = logs.Logs(self)
        self.plans = plans.Plans(self)
        self.reports = reports.Reports(self)
        self.roles = roles.Roles(self)
        self.settings = settings.Settings(self)
        self.storage_classes = storage_classes.StorageClasses(self)
        self.tasks = taskmgr.Tasks(self)
        self.templates = templates.Templates(self)
        self.users = users.Users(self)

    @property
    def ctera(self):
        return self.clients.ctera

    @property
    def api(self):
        return self.clients.api

    @property
    def io(self):
        return self.clients.io

    @property
    @abstractmethod
    def context(self):
        return NotImplementedError("Subclass must implement the 'context' property")

    @property
    def _session_id_key(self):
        return 'JSESSIONID'

    def _authenticator(self, url):
        return authenticators.core(self.session(), url, self.context)

    @property
    def _login_object(self):
        return login.Login(self)

    def test(self):
        return connection.test(self)

    def public_info(self):
        return self.ctera.get('/public/publicInfo')

    @property
    def _omit_fields(self):
        return super()._omit_fields + ['activation', 'admins', 'cloudfs', 'credentials', 'devices', 'directoryservice', 'domains', 'files',
                                       'firmwares', 'groups', 'logs', 'plans', 'reports', 'roles', 'settings', 'tasks', 'templates',
                                       'users']


class GlobalAdmin(Portal):  # pylint: disable=too-many-instance-attributes

    def __init__(self, host, port=None, https=True):
        super().__init__(host, port, https)
        self.antivirus = antivirus.Antivirus(self)
        self.buckets = buckets.Buckets(self)
        self.cli = cli.CLI(self)
        self.kms = kms.KMS(self)
        self.licenses = licenses.Licenses(self)
        self.mail = mail.Mail(self)
        self.messaging = messaging.Messaging(self)
        self.portals = portals.Portals(self)
        self.servers = servers.Servers(self)
        self.setup = setup.Setup(self)
        self.ssl = ssl.SSL(self)
        self.startup = startup.Startup(self)
        self.syslog = syslog.Syslog(self)

    @property
    def context(self):
        return 'admin'

    @property
    def _omit_fields(self):
        return super()._omit_fields + ['antivirus', 'buckets', 'cli', 'kms', 'licenses', 'mail', 'messaging', 'portals', 'servers',
                                       'setup', 'ssl', 'startup', 'syslog']


class ServicesPortal(Portal):

    @property
    def context(self):
        return 'ServicesPortal'



import logging
from .base_command import BaseCommand


class Mail(BaseCommand):
    """ Portal Mail Server Configuration APIs """

    def is_enabled(self):
        """
        Check if forwarding log messages over syslog is enabled
        """
        return self._core.api.get('/settings/enableEmailSending') is True

    def enable(self, host=None, port=None, sender=None, username=None, password=None, use_tls=None):
        """
        Enable Mail Server Notifications

        :param str,optional host: SMTP host
        :param str,optional port: SMTP port
        :param str,optional sender: Sender e-mail address
        :param str,optional username: The user name of the SMTP Server, defaults to None
        :param str,optional password: The password of the SMTP Server, defaults to None
        :param bool,optional use_tls: Use TLS when connecting to the SMTP Server, defaults to None
        """
        settings = self._core.api.get('/settings')
        settings.enableEmailSending = True
        settings = Mail._configure_settings(settings, host, port, sender, username, password, use_tls)
        logging.getLogger('cterasdk.core').info('Enabling Mail Server.')
        response = self._core.api.put('/settings', settings)
        logging.getLogger('cterasdk.core').info('Mail Server enabled.')
        return response

    def disable(self):
        """
        Disable Mail Server Notifications
        """
        logging.getLogger('cterasdk.core').info('Disabling Mail Server.')
        response = self._core.api.put('/settings/enableEmailSending', False)
        logging.getLogger('cterasdk.core').info('Mail Server disabled.')
        return response

    @staticmethod
    def _configure_settings(settings, host, port, sender, username, password, use_tls):
        if host is not None:
            settings.smtpSettings.smtpHost = host
        if port is not None:
            settings.smtpSettings.smtpPort = port
        if username is not None and password is not None:
            settings.smtpSettings.user = username
            settings.smtpSettings.password = password
        if use_tls is not None:
            settings.smtpSettings.enableTls = use_tls
        if sender is not None:
            settings.defaultPortalSettings.mailSettings.sender = sender
        return settings

    def modify(self, host=None, port=None, sender=None, username=None, password=None, use_tls=None):
        """
        Modify Mail Server Configuration

        :param str,optional host: SMTP host
        :param str,optional port: SMTP port
        :param str,optional sender: Sender e-mail address
        :param str,optional username: The user name of the SMTP Server, defaults to None
        :param str,optional password: The password of the SMTP Server, defaults to None
        :param bool,optional use_tls: Use TLS when connecting to the SMTP Server, defaults to None
        """
        settings = self._core.api.get('/settings')
        settings = Mail._configure_settings(settings, host, port, sender, username, password, use_tls)
        logging.getLogger('cterasdk.core').info('Modifying Mail Server settings.')
        response = self._core.api.put('/settings', settings)
        logging.getLogger('cterasdk.core').info('Mail Server settings modified.')
        return response



import logging

from .base_command import BaseCommand
from ..exceptions import CTERAException, ObjectNotFoundException
from ..common import Object, DateTimeUtils
from ..common import union
from . import query


class Administrators(BaseCommand):
    """
    Portal Global Administrators User Management APIs
    """

    default = ['name']

    def _get_entire_object(self, name):
        ref = f'/administrators/{name}'
        try:
            return self._core.api.get(ref)
        except CTERAException as error:
            raise CTERAException('Failed to get the user', error)

    def get(self, name, include=None):
        """
        Get a Global Administrator user account

        :param str name: Global administrator username
        :param list[str] include: List of fields to retrieve, defaults to ['name']
        :return: The user account, including the requested fields
        """
        baseurl = f'/administrators/{name}'
        include = union(include or [], Administrators.default)
        include = ['/' + attr for attr in include]
        user_object = self._core.api.get_multi(baseurl, include)
        if user_object.name is None:
            raise ObjectNotFoundException('Could not find user', baseurl, username=name)
        return user_object

    def list_admins(self, include=None):
        """
        List local administrators

        :param list[str] include: List of fields to retrieve, defaults to ['name']
        :return: Iterator for local administrators
        :rtype: cterasdk.lib.iterator.Iterator
        """
        include = union(include or [], Administrators.default)
        param = query.QueryParamBuilder().include(include).build()
        return query.iterator(self._core, '/administrators', param)

    def add(self, name, email, first_name, last_name, password, role, company=None, comment=None, password_change=False):
        """
        Create a Global Administrator

        :param str name: User name for the new GlobalAdmin
        :param str email: E-mail address of the new GlobalAdmin
        :param str first_name: The first name of the new GlobalAdmin
        :param str last_name: The last name of the new GlobalAdmin
        :param str password: Password for the new GlobalAdmin
        :param cterasdk.core.enum.Role role: User role of the new GlobalAdmin
        :param str,optional company: The name of the company of the new GlobalAdmin, defaults to None
        :param str,optional comment: Additional comment for the new GlobalAdmin, defaults to None
        :param variable,optional password_change:
            Require the user to change the password on the first login.
            Pass datetime.date for a specific date, integer for days from creation, or True for immediate , defaults to False
        """
        param = Object()
        param._classname = "PortalAdmin"  # pylint: disable=protected-access
        param.name = name
        param.email = email
        param.firstName = first_name
        param.lastName = last_name
        param.password = password
        param.role = role
        param.company = company
        param.comment = comment
        if password_change:
            param.requirePasswordChangeOn = DateTimeUtils.get_expiration_date(password_change).strftime('%Y-%m-%d')

        logging.getLogger('cterasdk.core').info('Creating a global administrator. %s', {'user': name})
        response = self._core.api.add('/administrators', param)
        logging.getLogger('cterasdk.core').info('Global administrator created. %s', {'user': name, 'email': email, 'role': role})

        return response

    def modify(self, current_username, new_username=None, email=None, first_name=None,
               last_name=None, password=None, role=None, company=None, comment=None):
        """
        Modify a Global Administrator user account

        :param str current_username: The current GlobalAdmin username
        :param str,optional new_username: New name
        :param str,optional email: E-mail address
        :param str,optional first_name: First name
        :param str,optional last_name: Last name
        :param str,optional password: Password
        :param cterasdk.core.enum.Role,optional role: User role
        :param str,optional company: Company name
        :param str,optional comment: Comment
        """
        user = self._get_entire_object(current_username)
        if new_username:
            user.name = new_username
        if email:
            user.email = email
        if first_name:
            user.firstName = first_name
        if last_name:
            user.lastName = last_name
        if password:
            user.password = password
        if role:
            user.role = role
        if company is not None:
            user.company = company
        if comment is not None:
            user.comment = comment

        try:
            response = self._core.api.put('/administrators/' + current_username, user)
            logging.getLogger('cterasdk.core').info("User modified. %s", {'username': user.name})
            return response
        except CTERAException as error:
            logging.getLogger('cterasdk.core').error("Failed to modify user.")
            raise CTERAException('Failed to modify user', error)

    def delete(self, name):
        """
        Delete a Global Administrator

        :param str username: Global administrator username
        """
        logging.getLogger('cterasdk.core').info('Deleting user. %s', {'user': name})
        baseurl = f'/administrators/{name}'
        response = self._core.api.execute(baseurl, 'delete', True)
        logging.getLogger('cterasdk.core').info('User deleted. %s', {'user': name})

        return response



from . import errors
from ..base import BaseClient, BaseResponse
from ..common import Serializers, Deserializers
from .. import async_requests, decorators
from ...common import Object


class AsyncClient(BaseClient):
    """Asynchronous Client"""

    @decorators.authenticated
    async def get(self, path, *, on_response=None, **kwargs):
        request = async_requests.GetRequest(self._builder(path), **kwargs)
        return await self.async_request(request, on_response=on_response)

    @decorators.authenticated
    async def put(self, path, data, *, data_serializer=None, on_response=None, **kwargs):
        request = async_requests.PutRequest(self._builder(path), data=data_serializer(data), **kwargs)
        return await self.async_request(request, on_response=on_response)

    @decorators.authenticated
    async def post(self, path, data, *, data_serializer=None, on_response=None, **kwargs):
        request = async_requests.PostRequest(self._builder(path), data=data_serializer(data), **kwargs)
        return await self.async_request(request, on_response=on_response)

    @decorators.authenticated
    async def form_data(self, path, data, *, on_response=None, **kwargs):
        request = async_requests.PostRequest(self._builder(path), data=Serializers.FormData(data), **kwargs)
        return await self.async_request(request, on_response=on_response)

    @decorators.authenticated
    async def delete(self, path, *, on_response=None, **kwargs):
        request = async_requests.DeleteRequest(self._builder(path), **kwargs)
        return await self.async_request(request, on_response=on_response)

    async def _request(self, request, *, on_response=None):
        on_response = on_response if on_response else AsyncResponse.new()
        response = await self._session.await_promise(self.join_headers(request), on_response=on_response)
        return await errors.accept(response)


class AsyncWebDAV(AsyncClient):
    """WebDAV"""


class AsyncJSON(AsyncClient):

    def __init__(self, builder=None, session=None, settings=None, authenticator=None):
        super().__init__(builder, session, settings, authenticator)
        self.headers.update_headers({'Content-Type': 'application/json'})

    async def get(self, path, **kwargs):
        response = await super().get(path, **kwargs)
        return await response.json()

    async def put(self, path, data, **kwargs):
        response = await super().put(path, data, data_serializer=Serializers.JSON, **kwargs)
        return await response.json()

    async def post(self, path, data, **kwargs):
        response = await super().post(path, data, data_serializer=Serializers.JSON, **kwargs)
        return await response.json()

    async def delete(self, path, **kwargs):
        response = await super().delete(path, **kwargs)
        return await response.json()


class AsyncXML(AsyncClient):

    async def get(self, path, **kwargs):
        response = await super().get(path, **kwargs)
        return await response.xml()

    async def put(self, path, data, **kwargs):
        response = await super().put(path, data, data_serializer=Serializers.XML, **kwargs)
        return await response.xml()

    async def post(self, path, data, **kwargs):
        response = await super().post(path, data, data_serializer=Serializers.XML, **kwargs)
        return await response.xml()

    async def delete(self, path, **kwargs):
        response = await super().delete(path, **kwargs)
        return await response.xml()


class AsyncExtended(AsyncXML):
    """CTERA Schema"""

    async def get_multi(self, path, paths):
        return await self.database(path, 'get-multi', paths)

    async def execute(self, path, name, param=None):  # schema method
        return await self._execute(path, 'user-defined', name, param)

    async def database(self, path, name, param=None):  # schema method
        return await self._execute(path, 'db', name, param)

    async def _execute(self, path, _type, name, param):
        data = Object()
        data.type = _type
        data.name = name
        data.param = param
        return await super().post(path, data)


class AsyncAPI(AsyncExtended):
    """CTERA Management API"""


class AsyncResponse(BaseResponse):
    """Asynchronous Response Object"""

    async def async_iter_content(self, chunk_size=None):
        async for chunk in self._response.content.iter_chunked(chunk_size if chunk_size else 5120):
            yield chunk

    async def text(self):
        return await self._response.text()

    async def json(self):
        return Deserializers.JSON(await self._response.read())

    async def xml(self):
        return Deserializers.XML(await self._response.read())

    async def read(self, n=-1):
        return await self._response.content.read(n)

    @staticmethod
    def new():
        async def new_response(response):
            return AsyncResponse(response)
        return new_response




import asyncio
import threading
from . import errors
from ..base import BaseClient
from ..common import Serializers
from ..asynchronous.clients import AsyncResponse
from ...common import Object
from .. import async_requests, decorators


class Client(BaseClient):
    """Synchronous Client"""

    @decorators.authenticated
    def handle(self, path, *, on_response=None, **kwargs):
        request = async_requests.GetRequest(self._builder(path), **kwargs)
        return self.request(request, on_response=on_response)

    @decorators.authenticated
    def get(self, path, *, on_response=None, **kwargs):
        request = async_requests.GetRequest(self._builder(path), **kwargs)
        return self.request(request, on_response=on_response)

    @decorators.authenticated
    def put(self, path, data, *, data_serializer=None, on_response=None, **kwargs):
        request = async_requests.PutRequest(self._builder(path), data=data_serializer(data), **kwargs)
        return self.request(request, on_response=on_response)

    @decorators.authenticated
    def post(self, path, data, *, data_serializer=None, on_response=None, **kwargs):
        request = async_requests.PostRequest(self._builder(path), data=data_serializer(data), **kwargs)
        return self.request(request, on_response=on_response)

    @decorators.authenticated
    def form_data(self, path, data, *, on_response=None, **kwargs):
        request = async_requests.PostRequest(self._builder(path), data=Serializers.FormData(data), **kwargs)
        return self.request(request, on_response=on_response)

    @decorators.authenticated
    def multipart(self, path, form, *, on_response=None, **kwargs):
        request = async_requests.PostRequest(self._builder(path), data=form.data, **kwargs)
        return self.request(request, on_response=on_response)

    @decorators.authenticated
    def delete(self, path, *, on_response=None, **kwargs):
        request = async_requests.DeleteRequest(self._builder(path), **kwargs)
        return self.request(request, on_response=on_response)

    def _request(self, request, *, on_response=None):
        on_response = on_response if on_response else SyncResponse.new()
        response = execute(self._session.await_promise, self.join_headers(request), on_response=on_response)
        return errors.accept(response)

    def close(self):  # pylint: disable=invalid-overridden-method
        return execute(super().close)


class Folders(Client):

    def download_zip(self, path, data, **kwargs):
        return super().form_data(path, data, **kwargs)


class Upload(Client):

    def upload(self, path, data, **kwargs):
        return super().form_data(path, data, **kwargs)


class WebDAV(Client):

    def download(self, path, **kwargs):
        return super().handle(path, **kwargs)

    def mkcol(self, path):
        request = async_requests.MkcolRequest(self._builder(path))
        response = self.request(request)
        return response.text()

    def copy(self, source, destination, *, overwrite=False):
        request = async_requests.CopyRequest(self._builder(source), headers=self._webdav_headers(destination, overwrite))
        response = self.request(request)
        return response.xml()

    def move(self, source, destination, *, overwrite=False):
        request = async_requests.MoveRequest(self._builder(source), headers=self._webdav_headers(destination, overwrite))
        response = self.request(request)
        return response.xml()

    def delete(self, path):  # pylint: disable=arguments-differ
        response = super().delete(path)
        return response.text()

    def _webdav_headers(self, destination, overwrite):
        return {
            'Destination': self._builder(destination),
            'Overwrite': 'T' if overwrite is True else 'F'
        }


class XML(Client):
    """XML Serializer and Deserializer"""

    def get(self, path, **kwargs):
        response = super().get(path, **kwargs)
        return response.xml()

    def put(self, path, data, **kwargs):
        response = super().put(path, data, data_serializer=Serializers.XML, **kwargs)
        return response.xml()

    def post(self, path, data, **kwargs):
        response = super().post(path, data, data_serializer=Serializers.XML, **kwargs)
        return response.xml()

    def form_data(self, path, data, **kwargs):
        response = super().form_data(path, data, **kwargs)
        return response.xml()

    def delete(self, path, **kwargs):
        response = super().delete(path, **kwargs)
        return response.xml()


class JSON(Client):
    """JSON Serializer and Deserializer"""

    def get(self, path, **kwargs):
        response = super().get(path, **kwargs)
        return response.json()

    def put(self, path, data, **kwargs):
        response = super().put(path, data, data_serializer=Serializers.JSON, **kwargs)
        return response.json()

    def post(self, path, data, **kwargs):
        response = super().post(path, data, data_serializer=Serializers.JSON, **kwargs)
        return response.json()

    def delete(self, path, **kwargs):
        response = super().delete(path, **kwargs)
        return response.json()


class Extended(XML):
    """CTERA Schema"""

    def get_multi(self, path, paths):
        return self.database(path, 'get-multi', paths)

    def show_multi(self, path, paths):
        print(Serializers.JSON(self.get_multi(path, paths), no_log=False))

    def execute(self, path, name, param=None):  # schema method
        return self._execute(path, 'user-defined', name, param)

    def database(self, path, name, param=None):  # schema method
        return self._execute(path, 'db', name, param)

    def add(self, path, param=None):
        return self.database(path, 'add', param)

    def _execute(self, path, _type, name, param):
        data = Object()
        data.type = _type
        data.name = name
        data.param = param
        return super().post(path, data)


class API(Extended):
    """CTERA Management API"""

    def defaults(self, classname):
        return self.get(f'/defaults/{classname}')


class Migrate(JSON):
    """CTERA Migrate Service"""

    ID = 'x-mt-x'

    def login(self):
        request = async_requests.GetRequest(self._builder('/auth/user'))
        response = self.request(request)
        token = response.headers.get(Migrate.ID, None)
        if token:
            self.headers.update_headers({Migrate.ID: token})
        return response.json()


event_loop = asyncio.new_event_loop()


def execute(target, *args, **kwargs):
    loop = asyncio.get_event_loop()

    if not loop.is_running():
        return loop.run_until_complete(target(*args, **kwargs))

    return run_threadsafe(event_loop, target, *args, **kwargs)


class SyncResponse(AsyncResponse):
    """Synchronous Response Object"""

    def iter_content(self, chunk_size=None):
        while True:
            try:
                yield execute(super().async_iter_content(chunk_size).__anext__)
            except StopAsyncIteration:
                break

    def text(self):  # pylint: disable=invalid-overridden-method
        return execute(super().text)

    def json(self):  # pylint: disable=invalid-overridden-method
        return execute(super().json)

    def xml(self):  # pylint: disable=invalid-overridden-method
        return execute(super().xml)

    @staticmethod
    def new():
        async def new_response(response):
            return SyncResponse(response)
        return new_response


def run_threadsafe(loop, target, *args, **kwargs):
    event = threading.Event()

    t = Task(loop, event, target, *args, **kwargs)
    t.start()

    event.wait()

    if t.exception:
        raise t.exception
    return t.response


class Task(threading.Thread):

    def __init__(self, loop, event, target, *args, **kwargs):
        super().__init__(name='Thread-safe Executor')
        self.loop = loop
        self.event = event
        self.target = target
        self.args = args
        self.kwargs = kwargs
        self.exception = None
        self.response = None

    def run(self):
        try:
            self.response = self.loop.run_until_complete(self.target(*self.args, **self.kwargs))
        except Exception as e:  # pylint: disable=broad-exception-caught
            self.exception = e
        finally:
            self.event.set()



import logging

from ..common import Object
from ..exceptions import CTERAException
from .base_command import BaseCommand


class Users(BaseCommand):
    """ Edge Filer Users configuration APIs """

    def get(self, name=None):
        """
        Get User. If a user name was not passed as an argument, a list of all local users will be retrieved

        :param str,optional name: Name of the user
        """
        return self._edge.api.get('/config/auth/users' + ('' if name is None else ('/' + name)))

    def add_first_user(self, username, password, email=''):
        """
        Add the first user of the Edge Filer and login

        :param str username: User name for the new user
        :param str password: Password for the new user
        :param str,optional email: E-mail address of the new user, defaults to an empty string
        """
        if not self._edge.initialized:
            user = Object()
            user.username = username
            user.password = password
            user.email = email
            self._edge.api.post('/nosession/createfirstuser', user)
            logging.getLogger('cterasdk.edge').info('User created. %s', {'user': username})
        else:
            logging.getLogger('cterasdk.edge').info('Skipping. root account already exists.')
        self._edge.login(username, password)

    def add(self, username, password, full_name=None, email=None, uid=None):
        """
        Add a user of the Gateway

        :param str username: User name for the new user
        :param str password: Password for the new user
        :param str,optional full_name: The full name of the new user, defaults to None
        :param str,optional email: E-mail address of the new user, defaults to None
        :param str,optional uid: The uid of the new user, defaults to None
        """
        user = Object()
        user.username = username
        user.password = password
        user.fullName = full_name
        user.email = email
        user.uid = uid
        try:
            response = self._edge.api.add('/config/auth/users', user)
            logging.getLogger('cterasdk.edge').info("User created. %s", {'username': user.username})
            return response
        except CTERAException as error:
            logging.getLogger('cterasdk.edge').error("User creation failed.")
            raise CTERAException('User creation failed', error)

    def modify(self, username, password=None, full_name=None, email=None, uid=None):
        """
        Modify an existing user of the Gateway

        :param str username: User name to modify
        :param str,optional password: New password, defaults to None
        :param str,optional full_name: The full name of the user, defaults to None
        :param str,optional email: E-mail address of the user, defaults to None
        :param str,optional uid: The uid of the user, defaults to None
        """
        try:
            user = self._edge.api.get('/config/auth/users/' + username)
        except CTERAException as error:
            raise CTERAException('Failed to get the user', error)

        if password:
            user.password = password
        if full_name:
            user.fullName = full_name
        if email:
            user.email = email
        if uid:
            user.uid = uid
        try:
            response = self._edge.api.put('/config/auth/users/' + username, user)
            logging.getLogger('cterasdk.edge').info("User modified. %s", {'username': user.username})
            return response
        except CTERAException as error:
            logging.getLogger('cterasdk.edge').error("Failed to modify user.")
            raise CTERAException('Failed to modify user', error)

    def delete(self, username):
        """
        Delete an existing user

        :param str username: User name of the user to delete
        """
        try:
            response = self._edge.api.delete('/config/auth/users/' + username)
            logging.getLogger('cterasdk.edge').info("User deleted. %s", {'username': username})
            return response
        except Exception as error:
            logging.getLogger('cterasdk.edge').error("User deletion failed.")
            raise CTERAException('User deletion failed', error)



import logging
from .exceptions import DirectIOAPIError, BlockError, StreamError


class Streamer:
    """
    Direct IO Streamer.
    """

    def __init__(self, executor, byte_range):
        """
        Initialize a Streamer.

        :param callable _executor: Asynchronous Direct IO Callable.
        :param cterasdk.direct.types.ByteRange byte_range: Byte Range.
        """
        self._executor = executor
        self._downloads = None
        self._byte_range = byte_range
        self._offset = byte_range.start

    def stop(self):
        """
        Stop Stream.
        """
        for download in self._downloads:
            download.cancel()

    async def start(self):
        """
        Stop Stream.
        """
        try:
            self._downloads = await self._executor()
            for download in self._downloads:
                block = await download
                fragment = block.fragment(self._byte_range)
                logging.getLogger('cterasdk.direct').debug('Streamer Fragment. %s', {'offset': fragment.offset, 'length': fragment.length})
                yield fragment
                self._offset = fragment.offset + fragment.length
        except DirectIOAPIError as error:
            raise StreamError(error.filename, self._offset)
        except BlockError as error:
            raise StreamError(error.block.file_id, self._offset)
        finally:
            self.stop()



import os
import sys
import logging
import cterasdk.settings

parameters = {
    'format': cterasdk.settings.logging.format,
    'datefmt': cterasdk.settings.logging.date_format,
}

filename = os.environ.get('cterasdk.log')
if filename:
    parameters['filename'] = filename
else:
    parameters['stream'] = sys.stdout

logging.basicConfig(**parameters)

for logger_conf in cterasdk.settings.logging.loggers:
    logger = logging.getLogger(f'{__package__}.{logger_conf.name}')
    logger.setLevel(logging.getLevelName(logger_conf.level.upper()))
